from xorq.common.utils import classproperty


try:
    from enum import StrEnum
except ImportError:
    from strenum import StrEnum


default_branch = "main"


class InitTemplates(StrEnum):
    cached_fetcher = "cached-fetcher"
    sklearn = "sklearn"
    penguins = "penguins"

    @classproperty
    def default(self):
        return self.cached_fetcher

    def get_default_branch(template, default_branch=default_branch):
        return dict(templates_branches).get(template, default_branch)


# NOTE: These are commit hashes from when the template update occurred
templates_branches = (
    (InitTemplates.cached_fetcher, "13696303173f138853cc77ae09f8e82a1e17afd4"),
    (InitTemplates.sklearn, "a915733da8fe69408e3254aa51e539017e0ac92a"),
    (InitTemplates.penguins, "034e12236e4935a62616253a7b096f7f29b92134"),
)



import contextlib
import io
import re
import shutil
import sys
import uuid
from itertools import chain
from pathlib import Path
from unittest.mock import patch

import pandas as pd
import pyarrow.parquet as pq
import pytest

import xorq.api as xo
from xorq.cli import (
    OutputFormats,
    build_command,
    run_command,
)
from xorq.common.utils.logging_utils import Run, Runs
from xorq.common.utils.node_utils import (
    find_node,
)
from xorq.common.utils.process_utils import (
    Popened,
    remove_ansi_escape,
    subprocess_run,
)
from xorq.flight.client import (
    FlightClient,
)
from xorq.ibis_yaml.compiler import (
    build_expr,
    load_expr,
)
from xorq.init_templates import InitTemplates


build_run_examples_expr_names = (
    ("local_cache.py", "expr"),
    ("multi_engine.py", "expr"),
    ("remote_caching.py", "expr"),
    ("iris_example.py", "expr"),
    ("simple_example.py", "expr"),
    ("deferred_read_csv.py", "pg_expr_replace"),
    ("train_test_splits.py", "train_table"),
    ("train_test_splits.py", "split_column"),
    ("postgres_caching.py", "expr"),
    ("xgboost_udaf.py", "expr"),
    ("expr_scalar_udf.py", "expr"),
    ("bank_marketing.py", "encoded_test"),
    ("flight_udtf_llm_example.py", "expr"),
    ("pyiceberg_backend_simple.py", "expr"),
    ("python_udwf.py", "expr"),
)


def test_build_command_function(tmp_path, fixture_dir):
    builds_dir = tmp_path / "builds"
    script_path = fixture_dir / "pipeline.py"

    build_command(script_path, "expr", str(builds_dir))
    assert builds_dir.exists()


def test_build_command(tmp_path, fixture_dir):
    builds_dir = tmp_path / "builds"
    script_path = fixture_dir / "pipeline.py"

    test_args = [
        "xorq",
        "build",
        str(script_path),
        "--expr-name",
        "expr",
        "--builds-dir",
        str(builds_dir),
    ]
    (returncode, _, stderr) = subprocess_run(test_args)

    assert "Building expr" in stderr.decode("ascii")
    assert returncode == 0, stderr
    assert builds_dir.exists()


@pytest.mark.slow(level=1)
def test_build_command_with_udtf(tmp_path, fixture_dir):
    builds_dir = tmp_path / "builds"
    script_path = fixture_dir / "udxf_expr.py"

    test_args = [
        "xorq",
        "build",
        str(script_path),
        "-e",
        "expr",
        "--builds-dir",
        str(builds_dir),
    ]
    (returncode, _, stderr) = subprocess_run(test_args)
    assert "Building expr" in stderr.decode("ascii")
    assert returncode == 0, stderr
    assert builds_dir.exists()


@pytest.mark.slow(level=1)
def test_build_command_on_notebook(monkeypatch, tmp_path, fixture_dir, capsys):
    builds_dir = tmp_path / "builds"
    script_path = fixture_dir / "pipeline.ipynb"

    test_args = [
        "xorq",
        "build",
        str(script_path),
        "-e",
        "expr",
        "--builds-dir",
        str(builds_dir),
    ]
    (returncode, _, stderr) = subprocess_run(test_args)

    assert "Building expr" in stderr.decode("ascii")
    assert returncode == 0, stderr
    assert builds_dir.exists()


@pytest.mark.slow(level=1)
def test_build_command_with_cache_dir(tmp_path, fixture_dir):
    builds_dir = tmp_path / "builds"
    cache_dir = tmp_path / "cache"
    script_path = fixture_dir / "pipeline.py"

    test_args = [
        "xorq",
        "build",
        str(script_path),
        "--expr-name",
        "expr",
        "--builds-dir",
        str(builds_dir),
        "--cache-dir",
        str(cache_dir),
    ]
    (returncode, _, stderr) = subprocess_run(test_args)

    assert "Building expr" in stderr.decode("ascii")
    assert returncode == 0, stderr
    assert builds_dir.exists()


def test_run_command_raises_on_unbound_expr(tmp_path):
    t = xo.table(schema={"a": "int64"})
    expr = t.filter(t.a > 0)
    build_dir = build_expr(expr, builds_dir=tmp_path)
    with pytest.raises(ValueError, match="Cannot run unbound expression"):
        run_command(build_dir)


@pytest.mark.slow(level=1)
def test_run_command_default(tmp_path, fixture_dir):
    target_dir = tmp_path / "build"
    script_path = fixture_dir / "pipeline.py"

    args = [
        "xorq",
        "build",
        str(script_path),
        "--expr-name",
        "expr",
        "--builds-dir",
        str(target_dir),
    ]
    (returncode, stdout, stderr) = subprocess_run(args)
    assert returncode == 0, stderr

    if match := re.search(f"{target_dir}/([0-9a-f]+)", stdout.decode("ascii")):
        expression_path = match.group()
        test_args = [
            "xorq",
            "run",
            expression_path,
        ]
        (returncode, _, _) = subprocess_run(test_args)

        # test with problematic name (see https://github.com/xorq-labs/xorq/issues/1116)
        test_args = [
            "xorq",
            "run",
            str(
                shutil.move(
                    expression_path,
                    Path(expression_path).parent.joinpath("becb4e71406b.bak"),
                )
            ),
        ]
        (returncode, _, stderr) = subprocess_run(test_args)

        assert returncode == 0, stderr
    else:
        raise AssertionError("No expression hash")


@pytest.mark.slow(level=1)
@pytest.mark.parametrize("output_format", OutputFormats)
def test_run_command(tmp_path, fixture_dir, output_format):
    target_dir = tmp_path / "build"
    script_path = fixture_dir / "pipeline.py"

    build_args = [
        "xorq",
        "build",
        str(script_path),
        "--expr-name",
        "expr",
        "--builds-dir",
        str(target_dir),
    ]
    (returncode, stdout, stderr) = subprocess_run(build_args)
    assert returncode == 0, stderr

    if match := re.search(f"{target_dir}/([0-9a-f]+)", stdout.decode("ascii")):
        output_path = tmp_path / f"test.{output_format}"
        expression_path = match.group()
        test_args = [
            "xorq",
            "run",
            expression_path,
            "--output-path",
            str(output_path),
            "--format",
            output_format,
        ]
        (returncode, _, stderr) = subprocess_run(test_args)
        assert returncode == 0, stderr
        assert output_path.exists()
        assert output_path.stat().st_size > 0

    else:
        raise AssertionError("No expression hash")


@pytest.mark.slow(level=1)
@pytest.mark.parametrize("output_format", ["csv", "json", "parquet"])
def test_run_cached_command(tmp_path, fixture_dir, output_format):
    target_dir = tmp_path / "build"
    script_path = fixture_dir / "pipeline.py"

    build_args = [
        "xorq",
        "build",
        str(script_path),
        "--expr-name",
        "expr",
        "--builds-dir",
        str(target_dir),
    ]
    (returncode, stdout, stderr) = subprocess_run(build_args)
    assert returncode == 0, stderr

    if match := re.search(f"{target_dir}/([0-9a-f]+)", stdout.decode("ascii")):
        output_path = tmp_path / f"test.{output_format}"
        expression_path = match.group()
        test_args = [
            "xorq",
            "run-cached",
            expression_path,
            "--output-path",
            str(output_path),
            "--format",
            output_format,
        ]
        (returncode, _, stderr) = subprocess_run(test_args)
        assert returncode == 0, stderr
        assert output_path.exists()
        assert output_path.stat().st_size > 0
    else:
        raise AssertionError("No expression hash")


@pytest.mark.slow(level=1)
def test_run_cached_command_default(tmp_path, fixture_dir):
    target_dir = tmp_path / "build"
    script_path = fixture_dir / "pipeline.py"

    build_args = [
        "xorq",
        "build",
        str(script_path),
        "--expr-name",
        "expr",
        "--builds-dir",
        str(target_dir),
    ]
    (returncode, stdout, stderr) = subprocess_run(build_args)
    assert returncode == 0, stderr

    if match := re.search(f"{target_dir}/([0-9a-f]+)", stdout.decode("ascii")):
        expression_path = match.group()
        test_args = [
            "xorq",
            "run-cached",
            expression_path,
        ]
        (returncode, _, stderr) = subprocess_run(test_args)
        assert returncode == 0, stderr
    else:
        raise AssertionError("No expression hash")


@pytest.mark.slow(level=1)
@pytest.mark.parametrize(
    "cache_type,ttl",
    [
        ("modification-time", None),
        ("snapshot", None),
        ("snapshot", "3600"),
    ],
)
def test_run_cached_command_cache_types(tmp_path, fixture_dir, cache_type, ttl):
    target_dir = tmp_path / "build"
    cache_dir = tmp_path / "cache"
    script_path = fixture_dir / "pipeline.py"

    build_args = [
        "xorq",
        "build",
        str(script_path),
        "--expr-name",
        "expr",
        "--builds-dir",
        str(target_dir),
    ]
    (returncode, stdout, stderr) = subprocess_run(build_args)
    assert returncode == 0, stderr

    if match := re.search(f"{target_dir}/([0-9a-f]+)", stdout.decode("ascii")):
        output_path = tmp_path / "test.parquet"
        expression_path = match.group()
        test_args = [
            "xorq",
            "run-cached",
            expression_path,
            "--cache-dir",
            str(cache_dir),
            "--cache-type",
            cache_type,
            "--output-path",
            str(output_path),
        ]
        if ttl is not None:
            test_args.extend(["--ttl", ttl])
        (returncode, _, stderr) = subprocess_run(test_args)
        assert returncode == 0, stderr
        assert output_path.exists()
        assert output_path.stat().st_size > 0
    else:
        raise AssertionError("No expression hash")


@pytest.mark.slow(level=1)
@pytest.mark.xdist_group(name="serve")
@pytest.mark.parametrize(
    "host,port,cache_dir", [(None, None, None), ("localhost", None, "cache")]
)
def test_serve_command(tmp_path, fixture_dir, cache_dir, host, port):
    target_dir = tmp_path / "build"
    script_path = fixture_dir / "udxf_pipeline.py"

    build_args = [
        "xorq",
        "build",
        str(script_path),
        "--expr-name",
        "expr",
        "--builds-dir",
        str(target_dir),
    ]
    (return_code, stdout, stderr) = subprocess_run(build_args)
    assert return_code == 0, stderr

    if match := re.search(f"{target_dir}/([0-9a-f]+)", stdout.decode("ascii")):
        expression_path = match.group()

        optional_args = tuple(
            chain.from_iterable(
                (arg, value)
                for arg, value in (
                    ("--cache-dir", str(tmp_path / cache_dir) if cache_dir else None),
                    ("--host", host),
                    ("--port", port),
                )
                if value
            )
        )

        serve_args = ("xorq", "serve-flight-udxf", str(expression_path), *optional_args)

        serve_process = Popened(serve_args)
        port = peek_port(serve_process)

        flight_con = xo.flight.connect(host=host, port=int(port))
        assert (
            serve_process.popen.poll() is None
            and "diamonds_exchange_command" in flight_con.list_exchanges()
        )
        serve_process.popen.terminate()

    else:
        raise AssertionError("No expression hash")


@pytest.mark.slow(level=1)
@pytest.mark.parametrize("output_format", ["csv", "json", "parquet"])
def test_run_command_stdout(tmp_path, fixture_dir, output_format):
    target_dir = tmp_path / "build"
    script_path = fixture_dir / "pipeline.py"
    build_args = [
        "xorq",
        "build",
        str(script_path),
        "--expr-name",
        "expr",
        "--builds-dir",
        str(target_dir),
    ]
    (returncode, stdout, stderr) = subprocess_run(build_args)
    assert returncode == 0, stderr

    if match := re.search(f"{target_dir}/([0-9a-f]+)", stdout.decode("ascii")):
        expression_path = match.group()
        test_args = [
            "xorq",
            "run",
            expression_path,
            "--output-path",
            "-",
            "--format",
            output_format,
        ]
        (returncode, stdout, stderr) = subprocess_run(test_args)
        assert returncode == 0, stderr
        assert stdout
    else:
        raise AssertionError("No expression hash")


def test_run_command_logging(tmp_path):
    """Test that run_command emits expected structured log events with metrics."""
    expr = xo.memtable({"a": [1, 2, 3], "b": [4, 5, 6]}, name="test_table")
    expr_path = build_expr(
        expr, builds_dir=tmp_path / "builds", cache_dir=tmp_path / "cache"
    )
    output_path = tmp_path / "out.parquet"
    runs_dir = tmp_path / "runs"

    with patch(
        "xorq.common.utils.logging_utils.get_xorq_runs_dir", return_value=runs_dir
    ):
        run_command(str(expr_path), str(output_path), "parquet")

    expr_hash = expr_path.name
    run: Run = Runs(expr_dir=runs_dir / expr_hash).runs[0]
    events = run.read_events()
    by_name = {e["event"]: e for e in events}

    # run.start captures all input params
    assert "run.start" in by_name
    start = by_name["run.start"]
    assert start["expr_path"] == str(expr_path)
    assert start["output_path"] == str(output_path)
    assert "output_format" in start
    assert "limit" in start

    # run.expr_loaded has elapsed timing
    assert "run.expr_loaded" in by_name
    assert by_name["run.expr_loaded"]["elapsed_s"] >= 0

    # run.done carries timing + output_format
    assert "run.done" in by_name
    done = by_name["run.done"]
    assert "elapsed_s" in done
    assert "output_format" in done

    # run.output_written carries file metrics for parquet output
    assert "run.output_written" in by_name
    written = by_name["run.output_written"]
    assert "bytes" in written
    assert written["bytes"] > 0
    assert "rows" in written
    assert written["rows"] == 3
    # rows uses pq.read_metadata (parquet footer) — verify it matches the actual file
    assert pq.read_metadata(output_path).num_rows == written["rows"]


def test_run_command_error_logging(tmp_path):
    """Test that run_command records error status in meta and re-raises exceptions."""
    runs_dir = tmp_path / "runs"
    nonexistent_path = tmp_path / "does_not_exist"
    expr_hash = nonexistent_path.name

    with patch(
        "xorq.common.utils.logging_utils.get_xorq_runs_dir", return_value=runs_dir
    ):
        with pytest.raises(Exception):  # noqa: B017
            run_command(str(nonexistent_path), str(tmp_path / "out.parquet"))

    run: Run = Runs(expr_dir=runs_dir / expr_hash).runs[0]
    meta = run.read_meta()
    assert meta["status"] == "error"
    assert "error" in meta


def test_run_command_writes_run_logger(tmp_path):
    expr = xo.memtable({"x": [10, 20, 30]}, name="t")
    expr_path = build_expr(
        expr, builds_dir=tmp_path / "builds", cache_dir=tmp_path / "cache"
    )
    output_path = tmp_path / "out.parquet"
    runs_dir = tmp_path / "runs"

    with patch(
        "xorq.common.utils.logging_utils.get_xorq_runs_dir", return_value=runs_dir
    ):
        run_command(str(expr_path), str(output_path), "parquet")

        expr_hash = expr_path.name
        runs_obj = Runs(expr_dir=runs_dir / expr_hash)
        assert len(runs_obj.list()) == 1, "Expected exactly one run to be recorded"

        run: Run = runs_obj.runs[0]
        assert uuid.UUID(run.run_id).version == 4, "Run ID should be a UUID4"

        meta = run.read_meta()
        assert meta is not None
        assert meta["status"] == "ok"
        assert meta["run_id"] == run.run_id
        assert meta["expr_hash"] == expr_hash
        assert meta["expr_path"] == str(expr_path)
        assert meta["output_format"] == "parquet"
        assert "started_at" in meta
        assert "completed_at" in meta

        events = run.read_events()
        event_names = [e["event"] for e in events]
        assert "run.start" in event_names
        assert "run.expr_loaded" in event_names
        assert "run.done" in event_names
        assert "run.output_written" in event_names

        loaded_event = next(e for e in events if e["event"] == "run.expr_loaded")
        assert "elapsed_s" in loaded_event
        assert loaded_event["elapsed_s"] >= 0

        written_event = next(e for e in events if e["event"] == "run.output_written")
        assert "bytes" in written_event
        assert written_event["bytes"] > 0


def test_run_command_run_logger_error_status(tmp_path):
    runs_dir = tmp_path / "runs"
    nonexistent_path = tmp_path / "does_not_exist"

    with patch(
        "xorq.common.utils.logging_utils.get_xorq_runs_dir", return_value=runs_dir
    ):
        with pytest.raises(Exception):  # noqa: B017
            run_command(str(nonexistent_path), str(tmp_path / "out.parquet"))

        expr_hash = nonexistent_path.name
        runs_obj = Runs(expr_dir=runs_dir / expr_hash)
        assert len(runs_obj.list()) == 1

        meta = runs_obj.runs[0].read_meta()
        assert meta is not None
        assert meta["status"] == "error"
        assert "error" in meta


def test_run_logger_multiple_runs(tmp_path):
    expr = xo.memtable({"v": [1, 2]}, name="t2")
    expr_path = build_expr(
        expr, builds_dir=tmp_path / "builds", cache_dir=tmp_path / "cache"
    )
    output_path = tmp_path / "out.parquet"
    runs_dir = tmp_path / "runs"

    with patch(
        "xorq.common.utils.logging_utils.get_xorq_runs_dir", return_value=runs_dir
    ):
        run_command(str(expr_path), str(output_path), "parquet")
        run_command(str(expr_path), str(output_path), "parquet")

        expr_hash = expr_path.name
        run_ids = Runs(expr_dir=runs_dir / expr_hash).list()
        assert len(run_ids) == 2
        assert run_ids[0] != run_ids[1]


@pytest.mark.parametrize(
    "expression,message",
    [
        ("integer", "The object integer must be an instance of"),
        ("missing", "Expression missing not found"),
    ],
)
@pytest.mark.slow(level=1)
def test_build_command_bad_expr_name(tmp_path, fixture_dir, expression, message):
    builds_dir = tmp_path / "builds"
    script_path = fixture_dir / "pipeline.py"

    test_args = [
        "xorq",
        "build",
        str(script_path),
        "-e",
        expression,
        "--builds-dir",
        str(builds_dir),
    ]
    (returncode, _, stderr) = subprocess_run(test_args)
    assert returncode != 0
    assert message in stderr.decode("ascii")


@pytest.mark.parametrize(
    ("example", "expr_name"),
    build_run_examples_expr_names,
)
@pytest.mark.slow(level=2)
def test_examples(
    example,
    expr_name,
    examples_dir,
    tmp_path,
):
    # build
    builds_dir = tmp_path / "builds"
    example_path = examples_dir / example
    assert example_path.exists()
    build_args = (
        "xorq",
        "build",
        str(example_path),
        "--expr-name",
        expr_name,
        "--builds-dir",
        str(builds_dir),
    )
    print(" ".join(build_args), file=sys.stderr)
    (returncode, stdout, stderr) = subprocess_run(build_args)
    assert returncode == 0, stderr
    print(stderr.decode("ascii"), file=sys.stderr)
    expression_path = Path(stdout.decode("ascii").strip().split("\n")[-1])
    # debugging can capture stdout and result in spurious path of "."
    assert expression_path.name and expression_path.exists()

    # run
    output_format = "parquet"
    output_path = expression_path / f"test.{output_format}"
    assert not output_path.exists()
    run_args = (
        "xorq",
        "run",
        str(expression_path),
        "--format",
        output_format,
        "--output-path",
        str(output_path),
    )
    print(" ".join(run_args), file=sys.stderr)
    (returncode, stdout, stderr) = subprocess_run(run_args)
    assert returncode == 0, stderr
    print(stderr, file=sys.stderr)
    assert output_path.exists()


def test_init_command_default(tmpdir):
    path = Path(tmpdir).joinpath("xorq-template-default")
    init_args = (
        "xorq",
        "init",
        "--path",
        str(path),
    )
    print(" ".join(init_args), file=sys.stderr)
    (returncode, stdout, stderr) = subprocess_run(init_args)
    assert returncode == 0, stderr
    assert path.exists()
    assert path.joinpath("pyproject.toml").exists()


@pytest.mark.parametrize("template", InitTemplates)
def test_init_command_sklearn(template, tmpdir):
    path = Path(tmpdir).joinpath(f"xorq-template-{template}")
    init_args = (
        "xorq",
        "init",
        "--path",
        str(path),
        "--template",
        template,
    )
    print(" ".join(init_args), file=sys.stderr)
    (returncode, stdout, stderr) = subprocess_run(init_args)
    assert returncode == 0, stderr
    assert path.exists()
    assert path.joinpath("pyproject.toml").exists()


@pytest.mark.parametrize("template", InitTemplates)
def test_init_command_path_exists(template, tmpdir):
    path = Path(tmpdir).joinpath(f"xorq-template-{template}")
    init_args = (
        "xorq",
        "init",
        "--path",
        str(path),
        "--template",
        template,
    )
    print(" ".join(init_args), file=sys.stderr)
    path.mkdir()
    (returncode, stdout, stderr) = subprocess_run(init_args)
    assert returncode != 0


@pytest.mark.slow(level=2)
@pytest.mark.skipif(
    sys.version_info < (3, 11), reason="requirements.txt issues for python3.10"
)
@pytest.mark.parametrize("template", InitTemplates)
def test_init_uv_build_uv_run(template, tmpdir):
    tmpdir = Path(tmpdir)
    path = tmpdir.joinpath(f"xorq-template-{template}")
    init_args = (
        "xorq",
        "init",
        "--path",
        str(path),
        "--template",
        template,
    )
    print(" ".join(init_args), file=sys.stderr)
    (returncode, stdout, stderr) = subprocess_run(init_args)
    assert returncode == 0, stderr
    assert path.exists()
    assert path.joinpath("pyproject.toml").exists()
    assert path.joinpath("requirements.txt").exists()

    build_args = (
        "xorq",
        "uv-build",
        str(path.joinpath("expr.py")),
    )
    (returncode, stdout, stderr) = subprocess_run(build_args, do_decode=True)
    assert returncode == 0, stderr
    build_path = Path(stdout.strip().split("\n")[-1])
    assert build_path.exists()

    output_path = tmpdir.joinpath("output")
    run_args = (
        "xorq",
        "uv-run",
        "--output-path",
        str(output_path),
        str(build_path),
    )
    (returncode, stdout, stderr) = subprocess_run(run_args, do_decode=True)
    assert returncode == 0, stderr
    assert output_path.exists()


serve_hashes = (
    "1e5f8b2b262dd841f1db1bc19ac6abc2",  # batting, rel.Read
    "e210634ae2df82a56dd49d7cf7516d14",  # awards_players, rel.Read
    "edaa2a90ba8df1688e67a24edb03964e",  # left, ops.Filter
    "929550bdf9130ece976a83ae2cf285f8",  # right, ops.DropColumns
)


@pytest.fixture(scope="session")
def pipeline_https_build(tmp_path_factory, fixture_dir):
    builds_dir = tmp_path_factory.mktemp("builds")
    script_path = fixture_dir / "pipeline_https.py"

    build_args = [
        "xorq",
        "build",
        str(script_path),
        "--expr-name",
        "expr",
        "--builds-dir",
        str(builds_dir),
    ]
    (returncode, stdout, stderr) = subprocess_run(build_args, do_decode=True)

    assert "Building expr" in stderr
    assert returncode == 0, stderr
    assert builds_dir.exists()
    serve_dir = Path(stdout.strip())
    return serve_dir


def peek_port(popened, timeout=60):
    def do_match(buf):
        (*_, line) = remove_ansi_escape(buf.decode("ascii").strip()).rsplit("\n", 1)
        match = re.match(".*on grpc://localhost:(\\d+)$", line)
        return match

    popened.popen.poll()
    if popened.popen.returncode:
        raise Exception(popened.stderr)
    try:
        buf = popened.stdout_peeker.peek_line_until(do_match, timeout=timeout)
    except TimeoutError as e:
        popened.popen.terminate()
        raise Exception(popened.stderr) from e
    (as_string,) = do_match(buf).groups()
    port = int(as_string)
    return port


def hit_server(port, expr):
    client = FlightClient(port=port)
    (_, rbr) = client.do_exchange("default", expr)
    df = rbr.read_pandas()
    return df


@pytest.mark.slow(level=3)
@pytest.mark.parametrize("serve_hash", serve_hashes)
def test_serve_unbound_hash(serve_hash, pipeline_https_build):
    lookup = {
        "929550bdf9130ece976a83ae2cf285f8": "xorq.vendor.ibis.expr.operations.DropColumns",
        "edaa2a90ba8df1688e67a24edb03964e": "xorq.vendor.ibis.expr.operations.Filter",
    }
    expr = load_expr(pipeline_https_build)
    typ = lookup.get(serve_hash)
    subexpr = find_node(expr, hash=serve_hash, tag=None, typs=typ).to_expr()

    serve_args = (
        "xorq",
        "serve-unbound",
        str(pipeline_https_build),
        "--to_unbind_hash",
        serve_hash,
    ) + (("--typ", typ) if typ else ())
    serve_popened = Popened(serve_args, deferred=False)
    port = peek_port(serve_popened)
    actual = hit_server(port=port, expr=subexpr)
    expected = expr.execute()
    (actual, expected) = (
        df.sort_values(list(df.columns), ignore_index=True) for df in (actual, expected)
    )
    assert actual.equals(expected)

    serve_popened.popen.terminate()


serve_tags = (
    "read-batting",
    "read-players",
    "batting-filtered",
    "players-filtered",
    # this needs the fix for finding the correct source
    "joined",
)


@pytest.mark.slow(level=3)
@pytest.mark.parametrize("serve_tag", serve_tags)
def test_serve_unbound_tag(serve_tag, pipeline_https_build):
    expr = load_expr(pipeline_https_build)
    subexpr = find_node(expr, hash=None, tag=serve_tag).to_expr()

    serve_args = (
        "xorq",
        "serve-unbound",
        str(pipeline_https_build),
        "--to_unbind_tag",
        serve_tag,
    )
    serve_popened = Popened(serve_args, deferred=False)
    port = peek_port(serve_popened)
    actual = hit_server(port=port, expr=subexpr)
    expected = expr.execute()
    (actual, expected) = (
        df.sort_values(list(df.columns), ignore_index=True) for df in (actual, expected)
    )
    assert actual.equals(expected)

    serve_popened.popen.terminate()


@pytest.mark.slow(level=1)
def test_serve_unbound_tag_get_exchange(pipeline_https_build, parquet_dir):
    batting_url = "https://storage.googleapis.com/letsql-pins/batting/20240711T171118Z-431ef/batting.parquet"
    serve_tag = "read-batting"
    expr = load_expr(pipeline_https_build)

    serve_args = (
        "xorq",
        "serve-unbound",
        str(pipeline_https_build),
        "--to_unbind_tag",
        serve_tag,
    )
    serve_popened = Popened(serve_args, deferred=False)
    port = peek_port(serve_popened)

    flight_backend = xo.flight.connect(port=port)
    f = flight_backend.get_exchange("default")
    actual = xo.deferred_read_parquet(batting_url).pipe(f).execute()

    expected = expr.execute()
    (actual, expected) = (
        df.sort_values(list(df.columns), ignore_index=True) for df in (actual, expected)
    )
    assert actual.equals(expected)

    serve_popened.popen.terminate()


@pytest.mark.slow(level=1)
def test_serve_unbound_tag_get_exchange_udf(fixture_dir, tmp_path):
    df = pd.DataFrame([float(v) for v in range(10)], columns=["x"])

    serve_tag = "full"

    builds_dir = tmp_path / "builds"
    script_path = fixture_dir / "pipeline_pandas_udf.py"

    # Capture print output
    output = io.StringIO()

    with contextlib.redirect_stdout(output):
        build_command(script_path, "expr", str(builds_dir))

    serve_args = (
        "xorq",
        "serve-unbound",
        str(output.getvalue().strip()),
        "--to_unbind_tag",
        serve_tag,
    )
    serve_popened = Popened(serve_args, deferred=False)
    port = peek_port(serve_popened)

    flight_backend = xo.flight.connect(port=port)
    f = flight_backend.get_exchange("default")
    actual = xo.connect().register(df).select("x").pipe(f).execute()

    assert not actual.empty

    serve_popened.popen.terminate()


@pytest.mark.slow(level=3)
def test_serve_penguins_template(tmpdir, tmp_path):
    tmpdir = Path(tmpdir)
    path = tmpdir.joinpath("xorq-template-penguins")
    init_args = (
        "xorq",
        "init",
        "--path",
        str(path),
        "--template",
        "penguins",
    )

    (returncode, stdout, stderr) = subprocess_run(init_args)

    assert returncode == 0, stderr
    assert path.exists()
    assert path.joinpath("pyproject.toml").exists()
    assert path.joinpath("requirements.txt").exists()

    target_dir = tmp_path / "build"
    build_args = [
        "xorq",
        "build",
        str(path / "expr.py"),
        "--builds-dir",
        str(target_dir),
    ]
    (returncode, stdout, stderr) = subprocess_run(build_args)

    assert "Building expr" in stderr.decode("ascii")
    assert returncode == 0, stderr

    if match := re.search(f"{target_dir}/([0-9a-f]+)", stdout.decode("ascii")):
        serve_hash = "e4122ce163e8680c757d6a5ed665440a"  # RemoteTable (test split)

        serve_args = (
            "xorq",
            "serve-unbound",
            str(target_dir / match.group()),
            "--to_unbind_hash",
            serve_hash,
        )
        serve_popened = Popened(serve_args, deferred=False)
        port = peek_port(serve_popened)

        # Create sample penguin data using memtable instead of reading from URL
        sample_data = pd.DataFrame(
            {
                "bill_length_mm": [
                    39.1,
                    39.5,
                    40.3,
                    36.7,
                    39.3,
                    38.9,
                    39.2,
                    34.1,
                    42.0,
                    37.8,
                ],
                "bill_depth_mm": [
                    18.7,
                    17.4,
                    18.0,
                    19.3,
                    20.6,
                    17.8,
                    19.6,
                    18.1,
                    20.2,
                    17.1,
                ],
                "species": [
                    "Adelie",
                    "Adelie",
                    "Adelie",
                    "Adelie",
                    "Adelie",
                    "Chinstrap",
                    "Chinstrap",
                    "Chinstrap",
                    "Gentoo",
                    "Gentoo",
                ],
            }
        )

        expr = xo.memtable(sample_data, name="penguins")

        actual = hit_server(port=port, expr=expr)
        assert not actual.empty
        assert actual["predict"].isin(("Adelie", "Chinstrap", "Gentoo")).all()
        assert len(actual) == len(sample_data)
    else:
        raise AssertionError("No expression hash")



import contextlib
import functools
import hashlib
import shutil
import zipfile
from pathlib import Path

import toolz
from attr import (
    field,
    frozen,
)
from attr.validators import (
    instance_of,
)


# paths that uv build injects that don't impact package functionality
uv_sdist_omit_suffixes = ("PKG-INFO", ".gitignore")


def copy_path(from_, to_):
    with from_.open("rb") as from_fh:
        with to_.open("wb") as to_fh:
            shutil.copyfileobj(from_fh, to_fh)


@toolz.curried.excepts(ValueError)
def try_path_relative_to(from_, to_):
    return Path(from_).relative_to(to_)


def uv_sdist_member_filter(name):
    return not name.endswith(uv_sdist_omit_suffixes)


def get_root_dir(zip_path):
    with zipfile.ZipFile(zip_path, "r") as zf:
        (name, *rest) = zf.namelist()
        (root_dir, *_) = Path(name).parts
        assert root_dir
        assert all(n.startswith(root_dir) for n in rest)
        return Path(root_dir)


def calc_zip_content_hexdigest(path, member_filter=uv_sdist_member_filter):
    from xorq.common.utils.dask_normalize.dask_normalize_utils import (  # noqa: PLC0415
        file_digest,
    )

    with zipfile.ZipFile(path, "r") as zf:
        names = sorted(
            name
            for name in zf.namelist()
            if not name.endswith("/") and member_filter(name)
        )
        md5 = hashlib.md5()
        for name in names:
            md5.update(file_digest(zf.open(name), hashlib.md5).encode("ascii"))
        return md5.hexdigest()


@frozen
class ZipProxy:
    zip_path = field(validator=instance_of(Path), converter=Path)

    def __attrs_post_init__(self):
        assert self.zip_path.suffix == ".zip"

    @functools.cached_property
    def root_dir(self):
        return get_root_dir(self.zip_path)

    def toplevel_name_exists(self, name):
        with self.open() as zf:
            return any(
                other and str(other) == name
                for other in (
                    try_path_relative_to(other, self.root_dir)
                    for other in zf.namelist()
                )
            )

    @contextlib.contextmanager
    def open(self):
        with zipfile.ZipFile(self.zip_path, "r") as zf:
            yield zf

    @contextlib.contextmanager
    def open_member(self, member_path):
        with self.open() as zf:
            yield zf.open(str(member_path))

    @contextlib.contextmanager
    def open_toplevel_member(self, member_path):
        with self.open_member(self.root_dir.joinpath(member_path)) as fh:
            yield fh

    def extract_toplevel_name(self, name, dest):
        dest = Path(dest)
        with dest.open("wb") as ofh:
            with self.open_toplevel_member(name) as ifh:
                ofh.write(ifh.read())
        return dest

    @property
    def members(self):
        with self.open() as zf:
            return zf.namelist()

    def extract_toplevel(self, dest):
        dest = Path(dest).absolute()
        with self.open() as zf:
            for name in zf.namelist():
                if name.endswith("/"):
                    continue
                relpath = Path(name).relative_to(self.root_dir)
                destpath = dest.joinpath(relpath)
                destpath.parent.mkdir(exist_ok=True, parents=True)
                destpath.write_bytes(zf.read(name))
        return dest


@frozen
class ZipAppender:
    zip_path = field(validator=instance_of(Path), converter=Path)
    append_path = field(validator=instance_of(Path), converter=Path)
    arcname = field(validator=instance_of(str))

    @classmethod
    def append_toplevel(cls, zip_path, append_path, **kwargs):
        append_path = Path(append_path)
        arcname = str(ZipProxy(zip_path).root_dir.joinpath(append_path.name))
        self = cls(zip_path, append_path, arcname=arcname)
        with zipfile.ZipFile(self.zip_path, "a") as zf:
            zf.write(self.append_path, arcname=self.arcname)
        return zip_path


def tgz_to_zip(tgz_path, zip_path=None):
    """Convert a .tar.gz archive to a .zip archive."""
    import tarfile  # noqa: PLC0415

    tgz_path = Path(tgz_path)
    if zip_path is None:
        zip_path = tgz_path.with_suffix("").with_suffix(".zip")
    zip_path = Path(zip_path)

    with tarfile.TarFile.gzopen(tgz_path) as tf:
        with zipfile.ZipFile(zip_path, "w", compression=zipfile.ZIP_DEFLATED) as zf:
            for member in tf.getmembers():
                if member.isdir():
                    continue
                fh = tf.extractfile(member)
                if fh is None:
                    continue
                zf.writestr(member.name, fh.read())

    return zip_path


__all__ = [
    "ZipAppender",
    "ZipProxy",
    "calc_zip_content_hexdigest",
    "copy_path",
    "get_root_dir",
    "tgz_to_zip",
]



import operator
from numbers import Real

import numpy as np
import pandas as pd
import pytest

import xorq.api as xo
import xorq.expr.datatypes as dt
from xorq.common.utils.graph_utils import walk_nodes
from xorq.expr.ml.enums import ResponseMethod
from xorq.expr.ml.pipeline_lib import Pipeline, Step, make_estimator_typ
from xorq.expr.relations import Tag
from xorq.vendor.ibis.common.collections import FrozenOrderedDict
from xorq.vendor.ibis.expr.types import Expr


sklearn = pytest.importorskip("sklearn")

from xorq.expr.ml.sklearn_utils import ColumnRemapper  # noqa: E402


# sklearn submodule imports
KMeans = sklearn.cluster.KMeans
MiniBatchKMeans = sklearn.cluster.MiniBatchKMeans
RandomForestClassifier = sklearn.ensemble.RandomForestClassifier
SimpleImputer = sklearn.impute.SimpleImputer
LinearRegression = sklearn.linear_model.LinearRegression
LogisticRegression = sklearn.linear_model.LogisticRegression
accuracy_score = sklearn.metrics.accuracy_score
adjusted_rand_score = sklearn.metrics.adjusted_rand_score
f1_score = sklearn.metrics.f1_score
get_scorer = sklearn.metrics.get_scorer
get_scorer_names = sklearn.metrics.get_scorer_names
make_scorer = sklearn.metrics.make_scorer
r2_score = sklearn.metrics.r2_score
SklearnPipeline = sklearn.pipeline.Pipeline
ColumnTransformer = sklearn.compose.ColumnTransformer
SelectKBest = sklearn.feature_selection.SelectKBest
f_classif = sklearn.feature_selection.f_classif
FeatureUnion = sklearn.pipeline.FeatureUnion
OneHotEncoder = sklearn.preprocessing.OneHotEncoder
StandardScaler = sklearn.preprocessing.StandardScaler
BaseEstimator = sklearn.base.BaseEstimator
cluster = sklearn.cluster

TARGET = "target"
features = (feature0, feature1) = ("feature_0", "feature_1")

get_metadata = operator.attrgetter("metadata")


@pytest.fixture(scope="module")
def t():
    return xo.memtable(
        {
            feature0: [1, 2],
            feature1: [3, 4],
            TARGET: [0, 1],
        }
    )


step_typs = (
    sklearn.preprocessing.StandardScaler,
    sklearn.linear_model.LinearRegression,
)


@pytest.fixture(scope="module")
def sklearn_pipeline():
    sklearn_pipeline = sklearn.pipeline.make_pipeline(*(typ() for typ in step_typs))
    return sklearn_pipeline


@pytest.fixture(scope="module")
def xorq_pipeline(sklearn_pipeline):
    xorq_pipeline = xo.Pipeline.from_instance(sklearn_pipeline)
    return xorq_pipeline


@pytest.fixture(scope="module")
def fitted_xorq_pipeline(xorq_pipeline, t):
    return xorq_pipeline.fit(t, target=TARGET)


def test_infer_features(fitted_xorq_pipeline):
    assert all(
        features == step.features for step in fitted_xorq_pipeline.transform_steps
    )


@pytest.fixture(scope="module")
def all_tags(t, fitted_xorq_pipeline):
    expr = fitted_xorq_pipeline.predict(t)
    all_tags = walk_nodes((Tag,), expr)
    return all_tags


def test_recover_pipeline_from_transform_expr(t, xorq_pipeline, fitted_xorq_pipeline):
    expr = fitted_xorq_pipeline.transform(t)
    assert expr.ls.pipeline == xorq_pipeline


def test_recover_pipeline_from_predict_expr(t, xorq_pipeline, fitted_xorq_pipeline):
    expr = fitted_xorq_pipeline.predict(t)
    assert expr.ls.pipeline == xorq_pipeline


def test_fitted_pipeline_pipeline_property(xorq_pipeline, fitted_xorq_pipeline):
    assert fitted_xorq_pipeline.pipeline == xorq_pipeline


def test_all_tags(t, fitted_xorq_pipeline, all_tags):
    expr = fitted_xorq_pipeline.predict(t)
    actual = tuple(map(get_metadata, expr.ls.get_tags()))
    expected = tuple(map(get_metadata, all_tags))
    assert actual == expected


@pytest.mark.parametrize(
    "pairs",
    (
        (("tag", "FittedStep-transform"),),
        (("tag", "FittedStep-predict"),),
        (
            ("tag", "FittedStep-transform"),
            ("tag", "FittedStep-predict"),
        ),
    ),
)
def test_tagging_pipeline(pairs, t, fitted_xorq_pipeline):
    def contains_any_pairs(d, pairs=pairs):
        return set(pairs).intersection(d.items())

    def sort_and_tuplify(dcts):
        return tuple(sorted(tuple(sorted(dct.items())) for dct in dcts))

    actual = sort_and_tuplify(
        map(
            get_metadata,
            fitted_xorq_pipeline.predict(t).ls.get_tags(
                predicate=contains_any_pairs,
            ),
        )
    )
    expected = sort_and_tuplify(
        dct
        for dct in (
            fitted_step.get_tag_kwargs()
            for fitted_step in fitted_xorq_pipeline.fitted_steps
        )
        if contains_any_pairs(dct)
    )
    assert actual and actual == expected


def test_score_expr_returns_metric(t, fitted_xorq_pipeline):
    score_expr = fitted_xorq_pipeline.score_expr(t)
    result = score_expr.execute()
    assert isinstance(result, Real)


def test_fitted_step_transform_known_schema_unpacks():
    """Test FittedStep.transform unpacks struct columns for known schema."""

    t = xo.memtable({"a": [1.0, 2.0, 3.0], "b": [4.0, 5.0, 6.0]})
    step = xo.Step.from_instance_name(StandardScaler(), name="scaler")
    fitted = step.fit(t, features=("a", "b"))

    result = fitted.transform(t)
    df = result.execute()

    # Should have unpacked columns a and b, not a struct column
    assert "a" in df.columns
    assert "b" in df.columns
    assert "transformed" not in df.columns


def test_fitted_step_transform_kv_encoded_no_unpack():
    """Test FittedStep.transform keeps KV-encoded column without unpacking."""

    t = xo.memtable({"cat": ["x", "y", "x", "z"]})
    step = xo.Step.from_instance_name(OneHotEncoder(), name="ohe")
    fitted = step.fit(t, features=("cat",))

    result = fitted.transform(t)
    df = result.execute()

    # Should have KV-encoded column named "transformed"
    assert "transformed" in df.columns
    # Should not have unpacked category columns
    assert "cat_x" not in df.columns
    assert "cat_y" not in df.columns
    assert "cat_z" not in df.columns


def test_fitted_step_transform_retain_others_true():
    """Test FittedStep.transform retains other columns by default."""

    t = xo.memtable({"a": [1.0, 2.0], "b": [3.0, 4.0], "other": ["x", "y"]})
    step = xo.Step.from_instance_name(StandardScaler(), name="scaler")
    fitted = step.fit(t, features=("a", "b"))

    result = fitted.transform(t, retain_others=True)
    df = result.execute()

    # Should retain the "other" column
    assert "other" in df.columns
    assert df["other"].tolist() == ["x", "y"]


def test_fitted_step_transform_retain_others_false():
    """Test FittedStep.transform drops other columns when retain_others=False."""

    t = xo.memtable({"a": [1.0, 2.0], "b": [3.0, 4.0], "other": ["x", "y"]})
    step = xo.Step.from_instance_name(StandardScaler(), name="scaler")
    fitted = step.fit(t, features=("a", "b"))

    result = fitted.transform(t, retain_others=False)
    df = result.execute()

    # Should not retain the "other" column
    assert "other" not in df.columns


def test_pipeline_get_output_columns_known_schema_features_propagate():
    """Test Pipeline correctly propagates features for known schema transformers."""

    t = xo.memtable({"a": [1.0, 2.0, 3.0], "b": [4.0, 5.0, 6.0], "y": [0.0, 1.0, 0.0]})
    pipeline = xo.Pipeline.from_instance(
        sklearn.pipeline.make_pipeline(StandardScaler(), LinearRegression())
    )
    fitted = pipeline.fit(t, target="y")

    # The transform step should have features = ("a", "b") from the known schema
    transform_step = fitted.transform_steps[0]
    assert transform_step.structer.get_output_columns() == ("a", "b")


def test_pipeline_get_output_columns_kv_encoded_features_use_dest_col():
    """Test Pipeline correctly uses dest_col for KV-encoded transformers."""

    t = xo.memtable({"cat": ["a", "b", "a"], "y": [0.0, 1.0, 0.0]})
    pipeline = xo.Pipeline.from_instance(
        sklearn.pipeline.make_pipeline(OneHotEncoder(), LinearRegression())
    )
    fitted = pipeline.fit(t, target="y")

    # The transform step should have features = ("transformed",) for KV-encoded
    transform_step = fitted.transform_steps[0]
    assert transform_step.structer.get_output_columns("transformed") == ("transformed",)


def test_pipeline_get_output_columns_mixed_transform_steps():
    """Test Pipeline with multiple transform steps propagates features correctly."""

    t = xo.memtable({"a": [1.0, None, 3.0], "b": [4.0, 5.0, 6.0], "y": [0.0, 1.0, 0.0]})
    pipeline = xo.Pipeline.from_instance(
        sklearn.pipeline.make_pipeline(
            SimpleImputer(), StandardScaler(), LinearRegression()
        )
    )
    fitted = pipeline.fit(t, target="y")

    # Both transform steps should have known schema
    for transform_step in fitted.transform_steps:
        assert not transform_step.structer.is_kv_encoded

    # Prediction should work
    result = fitted.predict(t)
    assert result.execute() is not None


def test_deeply_nested_kv_encoded_pipeline():
    """Test depth-4 nested pipeline with KV-encoded ColumnTransformer.

    Pipeline structure:
    - ColumnTransformer (KV-encoded due to OneHotEncoder)
      - FeatureUnion
        - Pipeline (SimpleImputer -> StandardScaler)
        - Pipeline (SimpleImputer -> StandardScaler)
      - Pipeline (SimpleImputer -> OneHotEncoder)
    - SelectKBest
    - RandomForestClassifier
    """

    # Create sample data
    np.random.seed(42)
    n_samples = 100

    data = pd.DataFrame(
        {
            "age": np.random.randint(18, 80, n_samples).astype(float),
            "income": np.random.randint(20000, 150000, n_samples).astype(float),
            "credit_score": np.random.randint(300, 850, n_samples).astype(float),
            "years_employed": np.random.randint(0, 40, n_samples).astype(float),
            "education": np.random.choice(
                ["high_school", "bachelor", "master", "phd"], n_samples
            ),
            "employment_type": np.random.choice(
                ["full_time", "part_time", "contract", "self_employed"], n_samples
            ),
            "region": np.random.choice(["north", "south", "east", "west"], n_samples),
            "approved": np.random.randint(0, 2, n_samples),
        }
    )

    numeric_features = ["age", "income", "credit_score", "years_employed"]
    categorical_features = ["education", "employment_type", "region"]
    all_features = tuple(numeric_features + categorical_features)

    # Build nested sklearn pipeline
    scaled_pipeline = SklearnPipeline(
        [
            ("imputer", SimpleImputer(strategy="median")),
            ("scaler", StandardScaler()),
        ]
    )

    imputed_pipeline = SklearnPipeline(
        [
            ("imputer", SimpleImputer(strategy="mean")),
            ("scaler", StandardScaler()),
        ]
    )

    numeric_union = FeatureUnion(
        [
            ("scaled", scaled_pipeline),
            ("imputed", imputed_pipeline),
        ]
    )

    categorical_pipeline = SklearnPipeline(
        [
            ("imputer", SimpleImputer(strategy="constant", fill_value="missing")),
            (
                "encoder",
                OneHotEncoder(handle_unknown="ignore", sparse_output=False),
            ),
        ]
    )

    preprocessor = ColumnTransformer(
        [
            ("numeric", numeric_union, numeric_features),
            ("categorical", categorical_pipeline, categorical_features),
        ]
    )

    sklearn_pipe = SklearnPipeline(
        [
            ("preprocessor", preprocessor),
            ("selector", SelectKBest(f_classif, k=10)),
            (
                "classifier",
                RandomForestClassifier(n_estimators=50, random_state=42),
            ),
        ]
    )

    # Fit and predict with xorq
    expr = xo.memtable(data)
    xorq_pipeline = Pipeline.from_instance(sklearn_pipe)
    fitted_pipeline = xorq_pipeline.fit(expr, features=all_features, target="approved")
    predictions = fitted_pipeline.predict(expr).execute()

    # Fit and predict with sklearn
    X = data[list(all_features)]
    y = data["approved"]
    sklearn_pipe.fit(X, y)
    sklearn_preds = sklearn_pipe.predict(X)

    # Assert predictions match
    assert np.array_equal(predictions[ResponseMethod.PREDICT].values, sklearn_preds)


def test_deeply_nested_non_kv_pipeline():
    """Test depth-4 nested pipeline with all known-schema transformers.

    Pipeline structure:
    - ColumnTransformer (known schema - no KV-encoded children)
      - Pipeline (SimpleImputer -> StandardScaler -> Pipeline)
        - Pipeline (SimpleImputer -> StandardScaler)
      - Pipeline (SimpleImputer -> StandardScaler)
    - RandomForestClassifier
    """

    # Create sample data
    np.random.seed(42)
    n_samples = 100

    data = pd.DataFrame(
        {
            "age": np.random.randint(18, 80, n_samples).astype(float),
            "income": np.random.randint(20000, 150000, n_samples).astype(float),
            "credit_score": np.random.randint(300, 850, n_samples).astype(float),
            "years_employed": np.random.randint(0, 40, n_samples).astype(float),
            "debt_ratio": np.random.uniform(0, 1, n_samples),
            "savings": np.random.randint(0, 100000, n_samples).astype(float),
            "approved": np.random.randint(0, 2, n_samples),
        }
    )

    numeric_features_a = ["age", "income", "credit_score"]
    numeric_features_b = ["years_employed", "debt_ratio", "savings"]
    all_features = tuple(numeric_features_a + numeric_features_b)

    # Build nested sklearn pipeline (depth 4)
    inner_pipeline = SklearnPipeline(
        [
            ("imputer2", SimpleImputer(strategy="mean")),
            ("scaler2", StandardScaler()),
        ]
    )

    numeric_a_pipeline = SklearnPipeline(
        [
            ("imputer", SimpleImputer(strategy="median")),
            ("scaler", StandardScaler()),
            ("inner", inner_pipeline),
        ]
    )

    numeric_b_pipeline = SklearnPipeline(
        [
            ("imputer", SimpleImputer(strategy="median")),
            ("scaler", StandardScaler()),
        ]
    )

    preprocessor = ColumnTransformer(
        [
            ("numeric_a", numeric_a_pipeline, numeric_features_a),
            ("numeric_b", numeric_b_pipeline, numeric_features_b),
        ]
    )

    sklearn_pipe = SklearnPipeline(
        [
            ("preprocessor", preprocessor),
            (
                "classifier",
                RandomForestClassifier(n_estimators=50, random_state=42),
            ),
        ]
    )

    # Fit and predict with xorq
    expr = xo.memtable(data)
    xorq_pipeline = Pipeline.from_instance(sklearn_pipe)
    fitted_pipeline = xorq_pipeline.fit(expr, features=all_features, target="approved")
    predictions = fitted_pipeline.predict(expr).execute()

    # Fit and predict with sklearn
    X = data[list(all_features)]
    y = data["approved"]
    sklearn_pipe.fit(X, y)
    sklearn_preds = sklearn_pipe.predict(X)

    # Assert predictions match
    assert np.array_equal(predictions[ResponseMethod.PREDICT].values, sklearn_preds)


def _scorer_info():
    """Return [(name, module, response_method), ...] for every registered scorer."""

    return [
        (
            name,
            get_scorer(name)._score_func.__module__,
            get_scorer(name)._response_method,
        )
        for name in get_scorer_names()
    ]


def get_scorers_by_type():
    """Categorize all sklearn scorers by their type based on internal module path."""
    info = _scorer_info()

    proba = {
        name
        for name, _, response in info
        if isinstance(response, tuple)
        or response in (ResponseMethod.PREDICT_PROBA, ResponseMethod.DECISION_FUNCTION)
    }

    multilabel = {name for name, _, _ in info if name.endswith("_samples")}

    # Non-multilabel scorers categorized by module
    non_ml = [
        (name, module) for name, module, _ in info if not name.endswith("_samples")
    ]

    classification = {
        name
        for name, module in non_ml
        if "_classification" in module or "_ranking" in module or "_scorer" in module
    }

    regression = {name for name, module in non_ml if "_regression" in module}

    cluster = {name for name, module in non_ml if "cluster" in module}

    return FrozenOrderedDict(
        **{
            k: tuple(sorted(v))
            for k, v in (
                ("classification", classification),
                ("regression", regression),
                ("cluster", cluster),
                ("proba", proba),
                ("multilabel", multilabel),
            )
        }
    )


SCORERS_BY_TYPE = get_scorers_by_type()


@pytest.fixture
def scoring_data():
    """Generate dataset suitable for classification, regression, and clustering."""

    np.random.seed(42)
    n = 100
    return {
        "x1": np.random.randn(n).tolist(),
        "x2": np.random.randn(n).tolist(),
        "y_class": (np.random.randn(n) > 0).astype(int).tolist(),
        "y_reg": (
            np.abs(np.random.randn(n)) + 0.1
        ).tolist(),  # positive for log/deviance scorers
    }


@pytest.mark.parametrize("scorer_name", SCORERS_BY_TYPE["classification"])
def test_pipeline_scoring_match_sklearn_classifier_scorer(scoring_data, scorer_name):
    """Test classification scorers match sklearn."""

    X = np.array([scoring_data["x1"], scoring_data["x2"]]).T
    y = np.array(scoring_data["y_class"])

    sklearn_pipe = SklearnPipeline(
        [
            ("scaler", StandardScaler()),
            ("model", LogisticRegression(random_state=42, max_iter=1000)),
        ]
    )
    sklearn_pipe.fit(X, y)

    t = xo.memtable(scoring_data)
    xorq_pipeline = xo.Pipeline.from_instance(sklearn_pipe)
    fitted_xorq = xorq_pipeline.fit(t, features=("x1", "x2"), target="y_class")

    scorer = get_scorer(scorer_name)
    sklearn_score = scorer(sklearn_pipe, X, y)
    xorq_score = fitted_xorq.score(X, y, scorer=scorer_name)

    np.testing.assert_allclose(xorq_score, sklearn_score, rtol=1e-9, atol=1e-12)


@pytest.mark.parametrize("scorer_name", SCORERS_BY_TYPE["regression"])
def test_pipeline_scoring_match_sklearn_regressor_scorer(scoring_data, scorer_name):
    """Test regression scorers match sklearn."""

    X = np.array([scoring_data["x1"], scoring_data["x2"]]).T
    y = np.array(scoring_data["y_reg"])

    sklearn_pipe = SklearnPipeline(
        [("scaler", StandardScaler()), ("model", LinearRegression())]
    )
    sklearn_pipe.fit(X, y)

    t = xo.memtable(scoring_data)
    xorq_pipeline = xo.Pipeline.from_instance(sklearn_pipe)
    fitted_xorq = xorq_pipeline.fit(t, features=("x1", "x2"), target="y_reg")

    scorer = get_scorer(scorer_name)
    sklearn_score = scorer(sklearn_pipe, X, y)
    xorq_score = fitted_xorq.score(X, y, scorer=scorer_name)

    np.testing.assert_allclose(xorq_score, sklearn_score, rtol=1e-9, atol=1e-12)


@pytest.mark.parametrize("scorer_name", SCORERS_BY_TYPE["cluster"])
def test_pipeline_scoring_match_sklearn_cluster_scorer(scoring_data, scorer_name):
    """Test clustering scorers match sklearn."""

    X = np.array([scoring_data["x1"], scoring_data["x2"]]).T
    y = np.array(scoring_data["y_class"])

    sklearn_pipe = SklearnPipeline(
        [
            ("scaler", StandardScaler()),
            ("clusterer", KMeans(n_clusters=2, random_state=42, n_init=10)),
        ]
    )
    sklearn_pipe.fit(X, y)

    t = xo.memtable(scoring_data)
    xorq_pipeline = xo.Pipeline.from_instance(sklearn_pipe)
    fitted_xorq = xorq_pipeline.fit(t, features=("x1", "x2"), target="y_class")

    scorer = get_scorer(scorer_name)
    sklearn_score = scorer(sklearn_pipe, X, y)
    xorq_score = fitted_xorq.score(X, y, scorer=scorer_name)

    np.testing.assert_allclose(xorq_score, sklearn_score, rtol=1e-9, atol=1e-12)


@pytest.fixture
def fitted_classifier():
    """Fitted classifier pipeline."""

    np.random.seed(42)
    data = {"x1": [0.0, 1.0], "x2": [0.0, 1.0], "y": [0, 1]}
    X = np.array([data["x1"], data["x2"]]).T
    y = np.array(data["y"])

    sklearn_pipe = SklearnPipeline(
        [("scaler", StandardScaler()), ("model", LogisticRegression())]
    )
    sklearn_pipe.fit(X, y)

    t = xo.memtable(data)
    fitted = xo.Pipeline.from_instance(sklearn_pipe).fit(
        t, features=("x1", "x2"), target="y"
    )
    return fitted, sklearn_pipe, X, y, t


@pytest.fixture
def fitted_regressor():
    """Fitted regressor pipeline."""

    np.random.seed(42)
    data = {"x1": [0.0, 1.0], "x2": [0.0, 1.0], "y": [0.0, 1.0]}
    X = np.array([data["x1"], data["x2"]]).T
    y = np.array(data["y"])

    sklearn_pipe = SklearnPipeline(
        [("scaler", StandardScaler()), ("model", LinearRegression())]
    )
    sklearn_pipe.fit(X, y)

    t = xo.memtable(data)
    fitted = xo.Pipeline.from_instance(sklearn_pipe).fit(
        t, features=("x1", "x2"), target="y"
    )
    return fitted, sklearn_pipe, X, y, t


def test_score_expr_default_scorer_classifier_is_accuracy(fitted_classifier):
    """Test default scorer for classifier is accuracy_score."""

    fitted, *_ = fitted_classifier
    scorer = fitted._get_default_scorer()
    assert scorer._score_func is accuracy_score


def test_score_expr_default_scorer_regressor_is_r2(fitted_regressor):
    """Test default scorer for regressor is r2_score."""

    fitted, *_ = fitted_regressor
    scorer = fitted._get_default_scorer()
    assert scorer._score_func is r2_score


def test_score_expr_default_scorer_cluster_is_adjusted_rand():
    """Test default scorer for clustering is adjusted_rand_score."""

    t = xo.memtable({"x": [0.0, 1.0], "y": [0, 1]})
    fitted = xo.Pipeline.from_instance(
        sklearn.pipeline.make_pipeline(KMeans(n_clusters=2, n_init=1))
    ).fit(t, features=("x",), target="y")

    assert fitted._get_default_scorer()._score_func is adjusted_rand_score


def test_score_expr_string_scorer(fitted_classifier):
    """Test passing a scorer name string."""

    fitted, sklearn_pipe, X, y, _ = fitted_classifier

    xorq_score = fitted.score(X, y, scorer="f1")
    sklearn_score = get_scorer("f1")(sklearn_pipe, X, y)
    np.testing.assert_allclose(xorq_score, sklearn_score, rtol=1e-9)


def test_score_expr_callable_scorer(fitted_classifier):
    """Test passing a raw callable metric function."""

    fitted, sklearn_pipe, X, y, _ = fitted_classifier

    xorq_score = fitted.score(X, y, scorer=f1_score)
    sklearn_score = f1_score(y, sklearn_pipe.predict(X))
    np.testing.assert_allclose(xorq_score, sklearn_score, rtol=1e-9)


def test_score_expr_make_scorer_object(fitted_classifier):
    """Test passing a make_scorer object directly."""

    fitted, sklearn_pipe, X, y, _ = fitted_classifier

    scorer = make_scorer(f1_score)
    xorq_score = fitted.score(X, y, scorer=scorer)
    sklearn_score = scorer(sklearn_pipe, X, y)
    np.testing.assert_allclose(xorq_score, sklearn_score, rtol=1e-9)


def test_score_expr_returns_expression(fitted_classifier):
    """Test score_expr returns an ibis expression."""

    fitted, _, _, _, t = fitted_classifier
    expr = fitted.score_expr(t, scorer="accuracy")

    assert isinstance(expr, Expr)
    assert isinstance(expr.execute(), (int, float))


def test_score_expr_default_scorer_raises_for_unknown_model():
    """Test _get_default_scorer raises ValueError for unknown model type."""

    # Custom estimator that isn't a Classifier/Regressor/Cluster
    class CustomEstimator(BaseEstimator):
        return_type = dt.int64  # needed for xorq to handle predict

        def fit(self, X, y=None):
            return self

        def predict(self, X):
            return [0] * len(X)

    t = xo.memtable({"x": [0.0, 1.0], "y": [0, 1]})
    fitted = xo.Pipeline.from_instance(
        sklearn.pipeline.make_pipeline(CustomEstimator())
    ).fit(t, features=("x",), target="y")

    with pytest.raises(ValueError, match="Cannot determine default scorer"):
        fitted._get_default_scorer()


def test_step_from_fit_transform_creates_transform_type():
    """Step.from_fit_transform creates a type with transform (not predict)."""

    def my_fit(X, y=None):
        return np.mean(X, axis=0)

    def my_transform(model, X, y=None):
        return X - model

    step = Step.from_fit_transform(
        fit=my_fit,
        transform=my_transform,
        return_type=dt.Array(dt.float64),
        name="custom_transform",
    )

    # Verify the dynamically created type has correct attributes
    assert hasattr(step.instance, "transform")
    assert hasattr(step.instance, "fit")
    assert not hasattr(step.instance, "predict")
    assert step.instance.return_type == dt.Array(dt.float64)


def test_step_from_fit_predict_creates_predict_type():
    """Step.from_fit_predict creates a type with predict (not transform)."""

    def my_fit(X, y=None):
        return int(np.median(y))

    def my_predict(model, X, y=None):
        return np.full(len(X), model)

    step = Step.from_fit_predict(
        fit=my_fit,
        predict=my_predict,
        return_type=dt.int64,
        name="custom_predict",
    )

    assert hasattr(step.instance, "predict")
    assert hasattr(step.instance, "fit")
    assert not hasattr(step.instance, "transform")
    assert step.instance.return_type == dt.int64


def test_step_from_fit_make_estimator_typ_both_raises():
    """Passing both transform and predict raises ValueError."""

    with pytest.raises(ValueError):
        make_estimator_typ(
            fit=lambda X, y=None: None,
            return_type=dt.float64,
            transform=lambda m, X: X,
            predict=lambda m, X: X,
        )


def test_step_from_fit_make_estimator_typ_neither_raises():
    """Passing neither transform nor predict raises ValueError."""

    with pytest.raises(ValueError):
        make_estimator_typ(
            fit=lambda X, y=None: None,
            return_type=dt.float64,
        )


def test_step_from_fit_predict_end_to_end():
    """Step.from_fit_predict works end-to-end with dest_col."""

    def my_fit(X, y=None):
        return int(np.median(y))

    def my_predict(model, X, y=None):
        return np.full(len(X), model)

    step = Step.from_fit_predict(
        fit=my_fit,
        predict=my_predict,
        return_type=dt.int64,
        name="custom_predict",
    )

    t = xo.memtable({"a": [1.0, 2.0, 3.0], "b": [4.0, 5.0, 6.0], "y": [0, 1, 1]})
    fitted = step.fit(t, features=("a", "b"), target="y", dest_col="pred")
    result = fitted.predict(t)
    df = result.execute()
    assert df is not None
    assert len(df) == 3


def test_feature_importances_fitted_step():
    """FittedStep.feature_importances returns importances for tree models."""

    t = xo.memtable(
        {
            "a": np.random.randn(50).tolist(),
            "b": np.random.randn(50).tolist(),
            "y": (np.random.randn(50) > 0).astype(int).tolist(),
        }
    )

    step = xo.Step.from_instance_name(
        RandomForestClassifier(n_estimators=5, random_state=42),
        name="rf",
    )
    fitted = step.fit(t, features=("a", "b"), target="y")
    result = fitted.feature_importances(t)
    df = result.execute()

    assert df is not None
    assert "feature_importances" in df.columns
    importances = df["feature_importances"].iloc[0]
    assert len(importances) == 2  # two features
    assert all(isinstance(v, float) for v in importances)


def test_feature_importances_fitted_pipeline():
    """FittedPipeline.feature_importances returns importances through pipeline."""

    t = xo.memtable(
        {
            "a": np.random.randn(50).tolist(),
            "b": np.random.randn(50).tolist(),
            "y": (np.random.randn(50) > 0).astype(int).tolist(),
        }
    )

    sklearn_pipe = SklearnPipeline(
        [
            ("scaler", StandardScaler()),
            ("rf", RandomForestClassifier(n_estimators=5, random_state=42)),
        ]
    )
    fitted = Pipeline.from_instance(sklearn_pipe).fit(
        t, features=("a", "b"), target="y"
    )

    result = fitted.feature_importances(t)
    df = result.execute()

    assert df is not None
    assert "feature_importances" in df.columns
    importances = df["feature_importances"].iloc[0]
    assert len(importances) == 2


@pytest.fixture
def cluster_data():
    """Generate data with clear cluster structure."""

    np.random.seed(42)
    # Two well-separated clusters
    cluster1 = np.random.randn(10, 2) + [0, 0]
    cluster2 = np.random.randn(10, 2) + [10, 10]
    data = np.vstack([cluster1, cluster2])
    return {"num1": data[:, 0].tolist(), "num2": data[:, 1].tolist()}


@pytest.mark.parametrize(
    "clusterer_cls,clusterer_kwargs",
    [
        pytest.param(
            "KMeans",
            {"n_clusters": 2, "random_state": 42, "n_init": 10},
            id="KMeans",
        ),
        pytest.param(
            "MiniBatchKMeans",
            {"n_clusters": 2, "random_state": 42, "n_init": 10},
            id="MiniBatchKMeans",
        ),
        pytest.param(
            "BisectingKMeans",
            {"n_clusters": 2, "random_state": 42},
            id="BisectingKMeans",
        ),
        pytest.param(
            "Birch",
            {"n_clusters": 2},
            id="Birch",
        ),
        pytest.param(
            "MeanShift",
            {},
            id="MeanShift",
        ),
        pytest.param(
            "AffinityPropagation",
            {"random_state": 42},
            id="AffinityPropagation",
        ),
    ],
)
def test_clustering_predict_inductive(cluster_data, clusterer_cls, clusterer_kwargs):
    """Test that inductive clustering algorithms support predict."""

    t = xo.memtable(cluster_data)
    features = ("num1", "num2")

    ClustererClass = getattr(cluster, clusterer_cls)
    clusterer = ClustererClass(**clusterer_kwargs)

    # xorq predict
    step = xo.Step.from_instance_name(clusterer, name="clusterer")
    fitted = step.fit(t, features=features)
    result = fitted.predict(t)
    xorq_labels = result.execute()[ResponseMethod.PREDICT].values

    # sklearn predict
    X = np.array([cluster_data["num1"], cluster_data["num2"]]).T
    sklearn_clusterer = ClustererClass(**clusterer_kwargs)
    sklearn_clusterer.fit(X)
    sklearn_labels = sklearn_clusterer.predict(X)

    # Labels should match
    np.testing.assert_array_equal(xorq_labels, sklearn_labels)


@pytest.mark.parametrize(
    "clusterer_cls,clusterer_kwargs",
    [
        pytest.param(
            "DBSCAN",
            {"eps": 3, "min_samples": 2},
            id="DBSCAN",
        ),
        pytest.param(
            "HDBSCAN",
            {"min_samples": 2},
            id="HDBSCAN",
        ),
        pytest.param(
            "AgglomerativeClustering",
            {"n_clusters": 2},
            id="AgglomerativeClustering",
        ),
        pytest.param(
            "SpectralClustering",
            {"n_clusters": 2, "random_state": 42},
            id="SpectralClustering",
        ),
        pytest.param(
            "OPTICS",
            {"min_samples": 2},
            id="OPTICS",
        ),
    ],
)
def test_clustering_predict_transductive_rejected_at_fit(
    cluster_data, clusterer_cls, clusterer_kwargs
):
    """Test that transductive clustering algorithms are rejected at fit time."""

    t = xo.memtable(cluster_data)
    features = ("num1", "num2")

    ClustererClass = getattr(cluster, clusterer_cls)
    clusterer = ClustererClass(**clusterer_kwargs)

    step = xo.Step.from_instance_name(clusterer, name="clusterer")

    with pytest.raises(ValueError, match="must have transform or predict method"):
        step.fit(t, features=features)


# ---------------------------------------------------------------------------
# remap_columns / remap_params
# ---------------------------------------------------------------------------

Ridge = sklearn.linear_model.Ridge
SimpleImputer = sklearn.impute.SimpleImputer


@pytest.fixture
def ct_pipeline():
    """Pipeline with a ColumnTransformer having num and cat slots."""
    numeric_transformer = SklearnPipeline(
        [
            ("imputer", SimpleImputer(strategy="median")),
            ("scaler", StandardScaler()),
        ]
    )
    preprocessor = ColumnTransformer(
        [
            ("num", numeric_transformer, ["age", "fare"]),
            (
                "cat",
                OneHotEncoder(handle_unknown="ignore", sparse_output=False),
                ["embarked", "sex"],
            ),
        ]
    )
    sk = SklearnPipeline(
        [
            ("preprocessor", preprocessor),
            ("classifier", LogisticRegression(max_iter=1000, C=1.0)),
        ]
    )
    return Pipeline.from_instance(sk)


def test_remap_columns_basic(ct_pipeline):
    remapped = ct_pipeline.remap_columns(
        {
            "preprocessor/num": ["distance", "flight_time"],
            "preprocessor/cat": ["carrier", "origin"],
        }
    )
    refs = ColumnRemapper.list_column_refs(remapped.instance)
    assert refs["preprocessor/num"] == ["distance", "flight_time"]
    assert refs["preprocessor/cat"] == ["carrier", "origin"]


def test_remap_columns_original_unchanged(ct_pipeline):
    ct_pipeline.remap_columns(
        {"preprocessor/num": ["x", "y"], "preprocessor/cat": ["z", "w"]}
    )
    refs = ColumnRemapper.list_column_refs(ct_pipeline.instance)
    assert refs["preprocessor/num"] == ["age", "fare"]
    assert refs["preprocessor/cat"] == ["embarked", "sex"]


def test_remap_columns_returns_pipeline(ct_pipeline):
    remapped = ct_pipeline.remap_columns(
        {"preprocessor/num": ["a", "b"], "preprocessor/cat": ["c"]}
    )
    assert isinstance(remapped, Pipeline)


def test_remap_columns_unknown_key_raises(ct_pipeline):
    with pytest.raises(ValueError, match="column_map keys not found"):
        ct_pipeline.remap_columns({"preprocessor/typo": ["a", "b"]})


def test_remap_columns_strict_raises_on_unmapped(ct_pipeline):
    with pytest.raises(ValueError, match="pipeline paths not covered"):
        ct_pipeline.remap_columns(
            {"preprocessor/num": ["a", "b"]},
            strict=True,
        )


def test_remap_columns_strict_passes_when_all_mapped(ct_pipeline):
    remapped = ct_pipeline.remap_columns(
        {"preprocessor/num": ["a", "b"], "preprocessor/cat": ["c", "d"]},
        strict=True,
    )
    assert isinstance(remapped, Pipeline)


def test_remap_columns_nested_slot(ct_pipeline):
    refs = ColumnRemapper.list_column_refs(ct_pipeline.instance)
    assert "preprocessor/num/imputer" not in refs  # imputer has no cols list
    assert "preprocessor/num" in refs


def test_remap_params_leaf_param(ct_pipeline):
    remapped = ct_pipeline.remap_params({"classifier__C": 0.01})
    assert remapped.instance.get_params()["classifier__C"] == 0.01


def test_remap_params_nested_param(ct_pipeline):
    remapped = ct_pipeline.remap_params(
        {"preprocessor__num__imputer__strategy": "mean"}
    )
    assert (
        remapped.instance.get_params()["preprocessor__num__imputer__strategy"] == "mean"
    )


def test_remap_params_whole_step_replacement(ct_pipeline):
    remapped = ct_pipeline.remap_params({"classifier": Ridge(alpha=0.5)})
    assert type(remapped.instance.named_steps["classifier"]) is Ridge
    assert remapped.instance.named_steps["classifier"].alpha == 0.5


def test_remap_params_original_unchanged(ct_pipeline):
    ct_pipeline.remap_params({"classifier__C": 0.001})
    assert ct_pipeline.instance.get_params()["classifier__C"] == 1.0


def test_remap_params_returns_pipeline(ct_pipeline):
    remapped = ct_pipeline.remap_params({"classifier__C": 0.5})
    assert isinstance(remapped, Pipeline)


def test_remap_params_unknown_key_raises(ct_pipeline):
    with pytest.raises(ValueError, match="param_map keys not found"):
        ct_pipeline.remap_params({"classifier__typo": 99})



from __future__ import annotations

import pytest

import xorq.vendor.ibis.expr.datatypes as dt
from xorq.vendor.ibis.common.temporal import IntervalUnit
from xorq.vendor.ibis.expr.datatypes.parse import parse


@pytest.mark.parametrize(
    "text,expected",
    [
        # boolean
        ("boolean", dt.boolean),
        ("bool", dt.boolean),
        # signed integers
        ("int8", dt.int8),
        ("int16", dt.int16),
        ("int32", dt.int32),
        ("int64", dt.int64),
        ("int", dt.int64),
        # unsigned integers
        ("uint8", dt.uint8),
        ("uint16", dt.uint16),
        ("uint32", dt.uint32),
        ("uint64", dt.uint64),
        # floats
        ("float16", dt.float16),
        ("halffloat", dt.float16),
        ("float32", dt.float32),
        ("float64", dt.float64),
        ("double", dt.float64),
        ("float", dt.float64),
        # temporal
        ("date", dt.date),
        ("time", dt.time),
        ("timestamp", dt.Timestamp()),
        ("timestamp('UTC')", dt.Timestamp(timezone="UTC")),
        ("timestamp('UTC', 3)", dt.Timestamp(timezone="UTC", scale=3)),
        ("timestamp(6)", dt.Timestamp(scale=6)),
        # binary / null
        ("binary", dt.binary),
        ("bytes", dt.binary),
        ("null", dt.null),
        # string family — bare and aliases
        ("str", dt.string),
        ("string", dt.string),
        ("!string", dt.String(nullable=False)),
        ("varchar", dt.string),
        ("varchar(100)", dt.string),
        ("!varchar(100)", dt.String(nullable=False)),
        ("char(10)", dt.string),
        # string(N) — length qualifier carries no semantic meaning and is stripped
        ("string(50)", dt.string),
        ("!string(50)", dt.String(nullable=False)),
        ("string(255)", dt.string),
        ("!string(255)", dt.String(nullable=False)),
        # decimal
        ("decimal", dt.Decimal()),
        ("decimal(10, 2)", dt.Decimal(10, 2)),
        ("decimal(38, 9)", dt.Decimal(38, 9)),
        ("bignumeric", dt.Decimal(76, 38)),
        ("bigdecimal(5, 3)", dt.Decimal(5, 3)),
        # interval
        ("interval('s')", dt.Interval(unit=IntervalUnit.SECOND)),
        ("interval('Y')", dt.Interval(unit=IntervalUnit.YEAR)),
        ("interval('D')", dt.Interval(unit=IntervalUnit.DAY)),
        # composite
        ("array<int64>", dt.Array(dt.int64)),
        ("array<string>", dt.Array(dt.string)),
        ("map<string, int64>", dt.Map(dt.string, dt.int64)),
        ("struct<a: int64, b: string>", dt.Struct({"a": dt.int64, "b": dt.string})),
        # extension types
        ("json", dt.JSON(binary=False)),
        ("jsonb", dt.JSON(binary=True)),
        ("uuid", dt.uuid),
        ("inet", dt.inet),
        ("macaddr", dt.macaddr),
        # nullable modifier
        ("!int32", dt.Int32(nullable=False)),
        ("!boolean", dt.Boolean(nullable=False)),
        ("!date", dt.Date(nullable=False)),
        ("!decimal(10, 2)", dt.Decimal(10, 2, nullable=False)),
        ("!array<int64>", dt.Array(dt.int64, nullable=False)),
    ],
)
def test_parse(text, expected):
    assert parse(text) == expected



import pytest

import xorq.api as xo
from xorq.caching.strategy import SnapshotStrategy
from xorq.common.utils.node_utils import (
    compute_expr_hash,
    find_by_expr_hash,
)
from xorq.expr.relations import HashingTag, Tag
from xorq.ibis_yaml.common import RefEnum, RegistryEnum
from xorq.ibis_yaml.compiler import YamlExpressionTranslator
from xorq.vendor import ibis


@pytest.fixture
def t():
    return ibis.table(
        {"a": "int64", "b": "string"},
        name="test_table",
    )


@pytest.fixture
def compiler():
    return YamlExpressionTranslator()


def test_different_hashing_tag_metadata_produces_distinct_hashes(t):
    """Two expressions with same structure but different HashingTag dimensions produce distinct hashes."""
    expr_v1 = t.hashing_tag("v1")
    expr_v2 = t.hashing_tag("v2")

    strategy = SnapshotStrategy()
    hash_v1 = compute_expr_hash(expr_v1, strategy=strategy)
    hash_v2 = compute_expr_hash(expr_v2, strategy=strategy)

    assert hash_v1 != hash_v2


def test_hashing_tag_differs_from_plain_tag(t):
    """expr.hashing_tag('v1') hashes differently from expr.tag('v1')."""
    tagged = t.tag("v1")
    hashing_tagged = t.hashing_tag("v1")

    strategy = SnapshotStrategy()
    hash_tag = compute_expr_hash(tagged, strategy=strategy)
    hash_hashing_tag = compute_expr_hash(hashing_tagged, strategy=strategy)

    assert hash_tag != hash_hashing_tag


def test_plain_tag_hash_unchanged(t):
    """Plain tag hash is same as untagged (backward compat)."""
    tagged = t.tag("v1")

    strategy = SnapshotStrategy()
    hash_tagged = compute_expr_hash(tagged, strategy=strategy)
    hash_plain = compute_expr_hash(t, strategy=strategy)

    assert hash_tagged == hash_plain


def test_hashing_tag_is_tag_subclass(t):
    """HashingTag IS-A Tag."""
    expr = t.hashing_tag("v1")
    node = expr.op()
    assert isinstance(node, HashingTag)
    assert isinstance(node, Tag)


def test_find_by_expr_hash_works_for_hashing_tag():
    """find_by_expr_hash can locate HashingTag nodes."""
    batting = xo.examples.batting.fetch()
    ht = batting.hashing_tag("v1")

    strategy = SnapshotStrategy()
    ht_hash = compute_expr_hash(ht, strategy=strategy)

    found = find_by_expr_hash(ht, ht_hash, typs=(HashingTag,))
    assert found is not None
    assert isinstance(found, HashingTag)


def test_yaml_roundtrip_preserves_hashing_tag(t, compiler):
    """YAML roundtrip preserves HashingTag."""
    expr = t.hashing_tag("v1", extra="data")
    yaml_dict = compiler.to_yaml(expr)

    # Check the op type in the YAML
    node_ref = yaml_dict["expression"][RefEnum.node_ref]
    expression = yaml_dict["definitions"][RegistryEnum.nodes][node_ref]
    assert expression["op"] == "HashingTag"

    roundtrip_expr = compiler.from_yaml(yaml_dict)
    assert isinstance(roundtrip_expr.op(), HashingTag)
    assert roundtrip_expr.op().metadata == expr.op().metadata
    assert roundtrip_expr.schema() == expr.schema()


def test_yaml_snapshot_hash_stored(t, compiler):
    """snapshot_hash is stored after YAML serialization for HashingTag."""
    expr = t.hashing_tag("v1")
    yaml_dict = compiler.to_yaml(expr)

    node_ref = yaml_dict["expression"][RefEnum.node_ref]
    expression = yaml_dict["definitions"][RegistryEnum.nodes][node_ref]

    assert "snapshot_hash" in expression


def test_nested_tag_outer_stripped_inner_preserved(t):
    """expr.hashing_tag('a').tag('b') — outer Tag stripped, inner HashingTag preserved."""
    inner = t.hashing_tag("a")
    outer = inner.tag("b")

    strategy = SnapshotStrategy()
    # outer tag is stripped, so hash should equal inner hashing_tag hash
    hash_outer = compute_expr_hash(outer, strategy=strategy)
    hash_inner = compute_expr_hash(inner, strategy=strategy)

    assert hash_outer == hash_inner


def test_nested_hashing_tags_both_preserved(t):
    """expr.hashing_tag('a').hashing_tag('b') — both are preserved, distinct from single."""
    single = t.hashing_tag("a")
    double = t.hashing_tag("a").hashing_tag("b")

    strategy = SnapshotStrategy()
    hash_single = compute_expr_hash(single, strategy=strategy)
    hash_double = compute_expr_hash(double, strategy=strategy)

    assert hash_single != hash_double



import pytest

import xorq.api as xo


@pytest.fixture(scope="session")
def con():
    return xo.duckdb.connect()


@pytest.fixture(scope="session")
def orders(con):
    return con.create_table(
        "orders",
        schema={
            "o_orderkey": "int32",
            "o_orderpriority": "string",
            "o_custkey": "int32",
            "o_orderdate": "date",
        },
        overwrite=True,
    )


@pytest.fixture(scope="session")
def supplier(con):
    return con.create_table(
        "supplier",
        schema={"s_suppkey": "int32", "s_nationkey": "int32"},
        overwrite=True,
    )


@pytest.fixture(scope="session")
def lineitem(con):
    return con.create_table(
        "lineitem",
        schema={
            "l_orderkey": "int64",
            "l_suppkey": "int32",
            "l_shipdate": "date",
            "l_extendedprice": "decimal(15,2)",
            "l_discount": "decimal(15,2)",
        },
        overwrite=True,
    )


@pytest.fixture(scope="session")
def customer(con):
    return con.create_table(
        "customer",
        schema={"c_custkey": "int32", "c_nationkey": "int32"},
        overwrite=True,
    )


@pytest.fixture(scope="session")
def nation(con):
    return con.create_table(
        "nation",
        schema={"n_nationkey": "int32", "n_name": "string"},
        overwrite=True,
    )


# Minimal test mimicking h07 join chain with self-reference in projection
def test_minimal_joinchain_self_reference(
    compiler, con, orders, supplier, lineitem, customer, nation
):
    q = supplier.join(lineitem, supplier.s_suppkey == lineitem.l_suppkey)
    q = q.join(orders, orders.o_orderkey == lineitem.l_orderkey)
    q = q.join(customer, customer.c_custkey == orders.o_custkey)
    n1 = nation
    n2 = nation.view()
    q = q.join(n1, supplier.s_nationkey == n1.n_nationkey)
    q = q.join(n2, customer.c_nationkey == n2.n_nationkey)
    q = q.projection(
        {
            "supp_nation": n1.n_name,
            "cust_nation": n2.n_name,
            "l_shipdate": lineitem.l_shipdate,
            "l_extendedprice": lineitem.l_extendedprice,
            "l_discount": lineitem.l_discount,
        }
    )
    q = q.filter(
        (
            ((q.cust_nation == "FRANCE") & (q.supp_nation == "GERMANY"))
            | ((q.cust_nation == "GERMANY") & (q.supp_nation == "FRANCE"))
        )
    )

    yaml_dict = compiler.to_yaml(q)
    profiles = {con._profile.hash_name: con}
    q_roundtrip = compiler.from_yaml(yaml_dict, profiles)

    try:
        _ = q_roundtrip["cust_nation"]
    except Exception as e:
        pytest.fail(f"Accessing 'cust_nation' on the roundtrip query failed: {e}")



"""GizmoSQL backend tests — single-backend and multi-engine."""

from __future__ import annotations

from operator import methodcaller
from pathlib import Path

import pandas as pd
import pyarrow as pa
import pytest

import xorq.api as xo
from xorq.caching import ParquetCache, SourceCache
from xorq.vendor.ibis import _


pytestmark = pytest.mark.gizmosql


# ── Single-backend tests ────────────────────────────────────────────────────


def test_connect(con):
    """Verify that the GizmoSQL connection is established."""
    assert con is not None
    assert con.name == "gizmosql"


def test_version(con):
    """Verify we can retrieve the server version."""
    version = con.version
    assert isinstance(version, str)
    assert len(version) > 0


def test_list_tables(con):
    """Verify that loaded test tables are listed."""
    tables = con.list_tables()
    assert isinstance(tables, list)
    assert "functional_alltypes" in tables
    assert "batting" in tables


def test_list_catalogs(con):
    """Verify that catalogs can be listed."""
    catalogs = con.list_catalogs()
    assert isinstance(catalogs, list)
    assert len(catalogs) > 0


def test_list_databases(con):
    """Verify that databases (schemas) can be listed."""
    databases = con.list_databases()
    assert isinstance(databases, list)
    assert "main" in databases


def test_current_catalog(con):
    """Verify current_catalog property works."""
    catalog = con.current_catalog
    assert isinstance(catalog, str)
    assert len(catalog) > 0


def test_current_database(con):
    """Verify current_database property works."""
    database = con.current_database
    assert isinstance(database, str)
    assert database == "main"


def test_execute(con, alltypes):
    """Verify executing an expression returns a pandas DataFrame."""
    result = alltypes.limit(10).execute()
    assert isinstance(result, pd.DataFrame)
    assert len(result) == 10


def test_to_pyarrow(con, alltypes):
    """Verify to_pyarrow returns a PyArrow Table."""
    result = con.to_pyarrow(alltypes.limit(10))
    assert isinstance(result, pa.Table)
    assert result.num_rows == 10


def test_to_pyarrow_batches(con, alltypes):
    """Verify to_pyarrow_batches returns a RecordBatchReader."""
    reader = con.to_pyarrow_batches(alltypes.limit(10))
    assert isinstance(reader, pa.ipc.RecordBatchReader)
    table = reader.read_all()
    assert table.num_rows == 10


def test_create_and_drop_table(con, temp_table):
    """Verify table creation and deletion via DDL."""
    t = xo.memtable({"a": [1, 2, 3], "b": ["x", "y", "z"]})
    con.create_table(temp_table, t)

    assert temp_table in con.list_tables()

    result = con.table(temp_table).execute()
    assert len(result) == 3

    with con._safe_raw_sql(f'DROP TABLE IF EXISTS "{temp_table}"'):
        pass
    assert temp_table not in con.list_tables()


def test_read_parquet(con):
    """Verify read_parquet via _read_local_and_ingest."""
    root = Path(__file__).resolve().parents[5]
    parquet_path = root / "ci" / "ibis-testing-data" / "parquet" / "diamonds.parquet"
    if not parquet_path.exists():
        pytest.skip("Test data not available")

    t = con.read_parquet(parquet_path, table_name="test_diamonds_read")
    result = t.limit(5).execute()
    assert isinstance(result, pd.DataFrame)
    assert len(result) == 5

    # Clean up
    with con._safe_raw_sql('DROP TABLE IF EXISTS "test_diamonds_read"'):
        pass


def test_memtable(con):
    """Verify in-memory table registration via ADBC."""
    t = xo.memtable({"x": [10, 20, 30], "y": ["a", "b", "c"]})
    result = con.execute(t)
    assert isinstance(result, pd.DataFrame)
    assert len(result) == 3
    assert list(result["x"]) == [10, 20, 30]


def test_schema(con, alltypes):
    """Verify we can retrieve the schema of a table."""
    schema = alltypes.schema()
    assert len(schema) > 0
    assert "id" in schema


# ── into_backend tests ──────────────────────────────────────────────────────


@pytest.mark.parametrize("method", ["execute", "to_pyarrow", "to_pyarrow_batches"])
def test_into_backend_to_duckdb(con, batting, method):
    """GizmoSQL → DuckDB via into_backend, all output methods."""
    ddb_con = xo.duckdb.connect()
    expr = batting.limit(50).into_backend(ddb_con, "gz_batting")
    res = methodcaller(method)(expr)

    if isinstance(res, pa.RecordBatchReader):
        res = res.read_all()
    if isinstance(res, pa.Table):
        assert res.num_rows == 50
    else:
        assert len(res) == 50


@pytest.mark.parametrize("method", ["execute", "to_pyarrow", "to_pyarrow_batches"])
def test_into_backend_to_datafusion(con, batting, method):
    """GizmoSQL → DataFusion (xorq default) via into_backend."""
    xorq_con = xo.connect()
    expr = batting.limit(50).into_backend(xorq_con, "gz_batting")
    res = methodcaller(method)(expr)

    if isinstance(res, pa.RecordBatchReader):
        res = res.read_all()
    if isinstance(res, pa.Table):
        assert res.num_rows == 50
    else:
        assert len(res) == 50


def test_into_backend_join_on_duckdb(con, batting, awards_players):
    """Transfer two GizmoSQL tables to DuckDB, join there."""
    ddb_con = xo.duckdb.connect()

    batting_ddb = batting.filter(batting.yearID >= 2010).into_backend(
        ddb_con, "gz_batting"
    )
    awards_ddb = awards_players.into_backend(ddb_con, "gz_awards")

    expr = (
        batting_ddb.join(awards_ddb, predicates=["playerID", "yearID"])
        .limit(20)
        .select("playerID", "yearID", "awardID", "H")
    )

    res = expr.execute()
    assert isinstance(res, pd.DataFrame)
    assert 0 < len(res) <= 20


def test_into_backend_double_hop(con, batting):
    """GizmoSQL → DuckDB → DataFusion (two hops)."""
    ddb_con = xo.duckdb.connect()
    xorq_con = xo.connect()

    expr = (
        batting.filter(batting.yearID == 2015)
        .select("playerID", "yearID", "teamID", "H")
        .into_backend(ddb_con, "gz_hop1")
        .filter(_.H > 100)
        .into_backend(xorq_con, "gz_hop2")
        .limit(10)
    )

    res = expr.execute()
    assert isinstance(res, pd.DataFrame)
    assert 0 < len(res) <= 10


def test_into_backend_reverse_duckdb_to_gizmosql(con, batting):
    """DuckDB → GizmoSQL: push local data into the GizmoSQL server."""
    ddb_con = xo.duckdb.connect()
    local_table = ddb_con.create_table(
        "local_players",
        batting.limit(100).to_pyarrow(),
    )
    expr = local_table.into_backend(con, "ddb_to_gz")

    res = expr.execute()
    assert isinstance(res, pd.DataFrame)
    assert len(res) == 100


# ── Caching tests ───────────────────────────────────────────────────────────


def test_into_backend_source_cache(con, batting):
    """GizmoSQL → DataFusion with SourceCache."""
    xorq_con = xo.connect()

    expr = (
        batting.filter(batting.yearID >= 2014)
        .limit(50)
        .into_backend(xorq_con, "gz_cached_batting")
        .cache(SourceCache.from_kwargs(source=xorq_con))
    )

    res = expr.execute()
    assert isinstance(res, pd.DataFrame)
    assert 0 < len(res) <= 50


def test_into_backend_parquet_cache(con, batting, tmp_path):
    """GizmoSQL → DuckDB with ParquetCache."""
    ddb_con = xo.duckdb.connect()

    expr = (
        batting.filter(batting.yearID >= 2014)
        .limit(50)
        .into_backend(ddb_con, "gz_pq_batting")
        .cache(ParquetCache.from_kwargs(source=ddb_con, relative_path=tmp_path))
    )

    res = expr.execute()
    assert isinstance(res, pd.DataFrame)
    assert 0 < len(res) <= 50

    # second execution should hit cache
    res2 = expr.execute()
    assert len(res2) == len(res)


def test_into_backend_chained_caches(con, batting, tmp_path):
    """GizmoSQL → DataFusion [SourceCache] → DuckDB [ParquetCache]."""
    xorq_con = xo.connect()
    ddb_con = xo.duckdb.connect()

    expr = (
        batting.filter(batting.yearID >= 2014)
        .limit(50)
        .into_backend(xorq_con, "gz_chain_batting")
        .cache(SourceCache.from_kwargs(source=xorq_con))
        .into_backend(ddb_con)
        .select("playerID", "yearID", "teamID", "H")
        .cache(ParquetCache.from_kwargs(source=ddb_con, relative_path=tmp_path))
    )

    res = expr.execute()
    assert isinstance(res, pd.DataFrame)
    assert 0 < len(res) <= 50



import tempfile
import zipfile
from contextlib import contextmanager
from pathlib import Path

from xorq.catalog.constants import (
    PREFERRED_SUFFIX,
)
from xorq.ibis_yaml.enums import REQUIRED_ARCHIVE_NAMES


def test_zip(zip_path):
    with zipfile.ZipFile(zip_path, "r") as zf:
        names = {Path(name).name for name in zf.namelist() if not name.endswith("/")}
        missing = set(REQUIRED_ARCHIVE_NAMES).difference(names)
        assert not missing, missing


@contextmanager
def make_zip_context(build_dir):
    path_prefix = (build_dir := Path(build_dir)).parent
    with tempfile.TemporaryDirectory() as td:
        zip_path = Path(td).joinpath(build_dir.name).with_suffix(PREFERRED_SUFFIX)
        with zipfile.ZipFile(zip_path, "w", compression=zipfile.ZIP_DEFLATED) as zf:
            for p in build_dir.rglob("*"):
                if p.is_file():
                    zf.write(p, arcname=p.relative_to(path_prefix))
        yield zip_path


@contextmanager
def extract_build_zip_context(zip_path):
    with tempfile.TemporaryDirectory() as td:
        with zipfile.ZipFile(zip_path, "r") as zf:
            zf.extractall(td)
        (path, *rest) = Path(td).iterdir()
        assert not rest
        yield path


def write_zip(path, relpath_to_bytes):
    path = Path(path)
    with zipfile.ZipFile(path, "w", compression=zipfile.ZIP_DEFLATED) as zf:
        for relpath, byts in dict(relpath_to_bytes).items():
            zf.writestr(str(relpath), byts)
    return path



from __future__ import annotations

from xorq.vendor.ibis.formats.pandas import PandasData


class DuckDBPandasData(PandasData):
    @staticmethod
    def convert_Array(s, dtype, pandas_type):
        return s.replace(float("nan"), None)


try:
    from xorq.vendor.ibis.formats.pyarrow import PyArrowData

    class DuckDBPyArrowData(PyArrowData):
        @staticmethod
        def convert_scalar(scalar, dtype):
            import pyarrow as pa

            result = PyArrowData.convert_scalar(scalar, dtype)
            if isinstance(result, pa.Scalar) and result.as_py() is None:
                return pa.scalar(None, type=result.type)
            return result

        @staticmethod
        def convert_column(column, dtype):
            import pyarrow as pa

            result = PyArrowData.convert_column(column, dtype)
            if isinstance(result, pa.ChunkedArray):
                if result.null_count == len(result):
                    return pa.chunked_array(
                        [pa.array([None] * len(result), type=result.type)]
                    )
            return result

except ImportError:
    pass



"""Tests for GizmoSQL vendored backend coverage.

Covers methods and branches in the vendored backend that are not
exercised by test_basic.py (single-backend, into_backend, caching).
"""

from __future__ import annotations

import contextlib
import json
import tempfile
from pathlib import Path
from urllib.parse import ParseResult

import pandas as pd
import pyarrow as pa
import pytest

import xorq.api as xo
import xorq.common.exceptions as exc
import xorq.vendor.ibis.expr.datatypes as dt
from xorq.vendor.ibis.backends.gizmosql import Backend
from xorq.vendor.ibis.backends.gizmosql.converter import DuckDBPyArrowData


pytestmark = pytest.mark.gizmosql


# ── _Settings ───────────────────────────────────────────────────────────────


def test_settings_getitem(con):
    """_Settings.__getitem__ returns a setting value."""
    tz = con.settings["TimeZone"]
    assert isinstance(tz, str)
    assert tz == "UTC"


def test_settings_getitem_missing(con):
    """_Settings.__getitem__ raises KeyError for unknown settings."""
    with pytest.raises(KeyError):
        con.settings["__nonexistent_setting__"]


def test_settings_repr(con):
    """_Settings.__repr__ returns a non-empty string."""
    r = repr(con.settings)
    assert isinstance(r, str)
    assert len(r) > 0


# ── create_table edge cases ─────────────────────────────────────────────────


def test_create_table_schema_only(con, temp_table):
    """create_table with schema but no obj creates an empty table."""
    con.create_table(temp_table, schema={"a": "int64", "b": "string"})
    assert temp_table in con.list_tables()
    result = con.table(temp_table).execute()
    assert isinstance(result, pd.DataFrame)
    assert len(result) == 0
    assert list(result.columns) == ["a", "b"]


def test_create_table_overwrite(con, temp_table):
    """create_table with overwrite=True replaces existing table."""
    t1 = xo.memtable({"x": [1, 2]})
    con.create_table(temp_table, t1)
    assert con.table(temp_table).count().execute() == 2

    t2 = xo.memtable({"x": [10, 20, 30]})
    con.create_table(temp_table, t2, overwrite=True)
    assert con.table(temp_table).count().execute() == 3


def test_create_table_from_dict(con, temp_table):
    """create_table with a dict (non-Expr obj) is converted via memtable."""
    con.create_table(temp_table, {"col_a": [1, 2, 3], "col_b": ["x", "y", "z"]})
    result = con.table(temp_table).execute()
    assert len(result) == 3


def test_create_table_no_obj_no_schema_raises(con, temp_table):
    """create_table without obj or schema raises ValueError."""
    with pytest.raises(ValueError, match="Either `obj` or `schema`"):
        con.create_table(temp_table)


def test_create_table_null_type_raises(con, temp_table):
    """create_table with a NULL typed column raises XorqTypeError."""
    with pytest.raises(exc.XorqTypeError, match="NULL typed columns"):
        con.create_table(temp_table, schema={"x": "null"})


# ── create_database / drop_database ─────────────────────────────────────────


def test_create_and_drop_database(con):
    """create_database and drop_database manage schemas."""
    db_name = "test_temp_schema"
    try:
        con.create_database(db_name)
        assert db_name in con.list_databases()
    finally:
        with contextlib.suppress(Exception):
            con.drop_database(db_name)


def test_create_database_with_catalog_raises(con):
    """create_database with a catalog argument raises."""
    with pytest.raises(exc.UnsupportedOperationError):
        con.create_database("test_db", catalog="some_catalog")


def test_drop_database_with_wrong_catalog_raises(con):
    """drop_database with a different catalog raises."""
    with pytest.raises(exc.UnsupportedOperationError):
        con.drop_database("test_db", catalog="nonexistent_catalog")


# ── list_databases with catalog filter ──────────────────────────────────────


def test_list_databases_with_catalog(con):
    """list_databases with catalog= filters results."""
    dbs = con.list_databases(catalog="memory")
    assert isinstance(dbs, list)
    assert "main" in dbs


# ── get_schema error handling ───────────────────────────────────────────────


def test_get_schema_table_not_found(con):
    """get_schema raises TableNotFound for nonexistent tables."""
    with pytest.raises(exc.TableNotFound):
        con.get_schema("__this_table_does_not_exist__")


# ── _convert_kwargs ─────────────────────────────────────────────────────────


def test_convert_kwargs():
    """_convert_kwargs converts read_only string to bool."""
    kwargs = {"read_only": "true"}
    Backend._convert_kwargs(kwargs)
    assert kwargs["read_only"] is True

    kwargs = {"read_only": "False"}
    Backend._convert_kwargs(kwargs)
    assert kwargs["read_only"] is False


# ── read_csv ────────────────────────────────────────────────────────────────


def test_read_csv(con):
    """read_csv reads a local CSV file via _read_local_and_ingest."""
    with tempfile.NamedTemporaryFile(suffix=".csv", mode="w", delete=False) as f:
        f.write("id,name,value\n")
        f.write("1,alice,10.5\n")
        f.write("2,bob,20.3\n")
        f.write("3,charlie,30.1\n")
        csv_path = f.name

    try:
        t = con.read_csv(csv_path, table_name="test_read_csv")
        result = t.execute()
        assert isinstance(result, pd.DataFrame)
        assert len(result) == 3
        assert "name" in result.columns
    finally:
        with con._safe_raw_sql('DROP TABLE IF EXISTS "test_read_csv"'):
            pass
        Path(csv_path).unlink(missing_ok=True)


# ── read_json ───────────────────────────────────────────────────────────────


def test_read_json(con):
    """read_json reads newline-delimited JSON via _read_local_and_ingest."""
    with tempfile.NamedTemporaryFile(suffix=".json", mode="w", delete=False) as f:
        for row in [
            {"id": 1, "name": "alice"},
            {"id": 2, "name": "bob"},
            {"id": 3, "name": "charlie"},
        ]:
            f.write(json.dumps(row) + "\n")
        json_path = f.name

    try:
        t = con.read_json(json_path, table_name="test_read_json")
        result = t.execute()
        assert isinstance(result, pd.DataFrame)
        assert len(result) == 3
    finally:
        with con._safe_raw_sql('DROP TABLE IF EXISTS "test_read_json"'):
            pass
        Path(json_path).unlink(missing_ok=True)


# ── vendored execute (nested/null types) ────────────────────────────────────


def test_vendored_execute_with_nulls(con):
    """Vendored execute handles columns with null values."""
    t = xo.memtable({"a": [1, None, 3], "b": ["x", None, "z"]})
    result = Backend.execute(con, t)
    assert isinstance(result, pd.DataFrame)
    assert len(result) == 3


def test_vendored_execute_with_list_type(con):
    """Vendored execute handles nested (list) types via to_pylist()."""
    with con._safe_raw_sql(
        "CREATE OR REPLACE TABLE test_nested AS SELECT [1, 2, 3] AS arr, 'a' AS label"
    ):
        pass
    try:
        t = con.table("test_nested")
        result = Backend.execute(con, t)
        assert isinstance(result, pd.DataFrame)
        assert len(result) == 1
    finally:
        with con._safe_raw_sql('DROP TABLE IF EXISTS "test_nested"'):
            pass


# ── vendored to_pyarrow (column / scalar) ──────────────────────────────────


def test_vendored_to_pyarrow_column(con, batting):
    """Vendored to_pyarrow returns ChunkedArray for column expressions."""
    col_expr = batting.limit(5).playerID
    result = Backend.to_pyarrow(con, col_expr)
    assert isinstance(result, pa.ChunkedArray)
    assert len(result) == 5


def test_vendored_to_pyarrow_scalar(con, batting):
    """Vendored to_pyarrow returns a scalar for scalar expressions."""
    scalar_expr = batting.count()
    result = Backend.to_pyarrow(con, scalar_expr)
    assert isinstance(result, pa.Scalar) or isinstance(result, int)


# ── vendored to_pyarrow_batches ─────────────────────────────────────────────


def test_vendored_to_pyarrow_batches(con, batting):
    """Vendored to_pyarrow_batches returns RecordBatchReader."""
    reader = Backend.to_pyarrow_batches(con, batting.limit(10))
    assert isinstance(reader, pa.ipc.RecordBatchReader)
    table = reader.read_all()
    assert table.num_rows == 10


# ── _normalize_arrow_schema ─────────────────────────────────────────────────


def test_normalize_arrow_schema_large_types(con):
    """_normalize_arrow_schema downcasts large_string to string."""
    table = pa.table(
        {
            "a": pa.array(["x", "y", "z"], type=pa.large_string()),
            "b": pa.array([b"a", b"b", b"c"], type=pa.large_binary()),
            "c": pa.array([1, 2, 3], type=pa.int64()),
        }
    )
    result = con._normalize_arrow_schema(table)
    assert result.schema.field("a").type == pa.string()
    assert result.schema.field("b").type == pa.binary()
    assert result.schema.field("c").type == pa.int64()


def test_normalize_arrow_schema_no_change(con):
    """_normalize_arrow_schema returns table unchanged when no large types."""
    table = pa.table({"a": [1, 2, 3], "b": ["x", "y", "z"]})
    result = con._normalize_arrow_schema(table)
    assert result is table  # same object, no copy


# ── _register_in_memory_table (empty table) ─────────────────────────────────


def test_register_empty_memtable(con):
    """Empty memtable is created via DDL (ADBC ingest fails on zero rows)."""
    t = xo.memtable(
        pa.table(
            {"x": pa.array([], type=pa.int64()), "y": pa.array([], type=pa.string())}
        )
    )
    result = con.execute(t)
    assert isinstance(result, pd.DataFrame)
    assert len(result) == 0


# ── _get_schema_using_query ─────────────────────────────────────────────────


def test_get_schema_using_query(con):
    """_get_schema_using_query returns Schema from DESCRIBE."""
    schema = con._get_schema_using_query("SELECT 1 AS a, 'hello' AS b")
    assert "a" in schema
    assert "b" in schema


# ── _from_url ───────────────────────────────────────────────────────────────


def test_from_url_parsing(con):
    """_from_url parses URL components correctly."""
    url = ParseResult(
        scheme="gizmosql",
        netloc="myuser:mypass@myhost:31337",
        path="/mydb/myschema",
        params="",
        query="useEncryption=true&disableCertificateVerification=true",
        fragment="",
    )

    # We can't actually connect to this URL, but we can test the parsing
    # by mocking connect
    parsed_kwargs = {}

    original_connect = Backend.connect

    def mock_connect(self, **kwargs):
        parsed_kwargs.update(kwargs)
        raise ConnectionError("mock")

    Backend.connect = mock_connect
    try:
        backend = Backend()
        with pytest.raises(ConnectionError, match="mock"):
            backend._from_url(
                url, useEncryption="true", disableCertificateVerification="true"
            )
        assert parsed_kwargs["host"] == "myhost"
        assert parsed_kwargs["port"] == 31337
        assert parsed_kwargs["user"] == "myuser"
        assert parsed_kwargs["password"] == "mypass"
        assert parsed_kwargs["database"] == "mydb"
        assert parsed_kwargs["schema"] == "myschema"
        assert parsed_kwargs["use_encryption"] is True
        assert parsed_kwargs["disable_certificate_verification"] is True
    finally:
        Backend.connect = original_connect


# ── _create_temp_view / _get_temp_view_definition ───────────────────────────


def test_create_temp_view(con):
    """_create_temp_view creates a temporary view."""
    view_name = "test_temp_view_cov"
    try:
        con._create_temp_view(view_name, "SELECT 42 AS answer")
        with con._safe_raw_sql(f'SELECT * FROM "{view_name}"') as cur:
            result = cur.fetch_arrow_table()
        assert result.column("answer")[0].as_py() == 42
    finally:
        with contextlib.suppress(Exception):
            with con._safe_raw_sql(f'DROP VIEW IF EXISTS "{view_name}"'):
                pass


# ── from_connection ─────────────────────────────────────────────────────────


def test_from_connection(con):
    """from_connection creates a new backend from an existing ADBC connection."""
    new_backend = Backend.from_connection(con.con)
    assert new_backend.current_database == "main"
    tables = new_backend.list_tables()
    assert isinstance(tables, list)


# ── _from_url OAuth kwargs ───────────────────────────────────────────────────


def test_from_url_oauth_kwargs():
    """_from_url converts camelCase OAuth kwargs to snake_case."""
    parsed_kwargs = {}
    original_connect = Backend.connect

    def mock_connect(self, **kwargs):
        parsed_kwargs.update(kwargs)
        raise ConnectionError("mock")

    Backend.connect = mock_connect
    try:
        backend = Backend()
        url = ParseResult(
            scheme="gizmosql",
            netloc="user:pass@host:31337",
            path="",
            params="",
            query="",
            fragment="",
        )
        with pytest.raises(ConnectionError, match="mock"):
            backend._from_url(
                url,
                authType="oauth",
                oauthPort="8080",
                oauthTimeout="120",
                openBrowser="false",
            )
        assert parsed_kwargs["auth_type"] == "oauth"
        assert parsed_kwargs["oauth_port"] == 8080
        assert parsed_kwargs["oauth_timeout"] == 120
        assert parsed_kwargs["open_browser"] is False
    finally:
        Backend.connect = original_connect


# ── DuckDBPyArrowData converter ──────────────────────────────────────────────


def test_pyarrow_data_convert_scalar_null():
    """DuckDBPyArrowData.convert_scalar handles null scalars."""
    result = DuckDBPyArrowData.convert_scalar(
        pa.scalar(None, type=pa.int64()), dt.int64
    )
    assert isinstance(result, pa.Scalar)
    assert result.as_py() is None


def test_pyarrow_data_convert_scalar_non_null():
    """DuckDBPyArrowData.convert_scalar passes through non-null scalars."""
    result = DuckDBPyArrowData.convert_scalar(pa.scalar(42, type=pa.int64()), dt.int64)
    assert result.as_py() == 42


def test_pyarrow_data_convert_column_all_null():
    """DuckDBPyArrowData.convert_column handles all-null columns."""
    col = pa.chunked_array([pa.array([None, None, None], type=pa.int64())])
    result = DuckDBPyArrowData.convert_column(col, dt.int64)
    assert isinstance(result, pa.ChunkedArray)
    assert len(result) == 3
    assert result.null_count == 3


def test_pyarrow_data_convert_column_non_null():
    """DuckDBPyArrowData.convert_column passes through non-null columns."""
    col = pa.chunked_array([pa.array([1, 2, 3], type=pa.int64())])
    result = DuckDBPyArrowData.convert_column(col, dt.int64)
    assert isinstance(result, pa.ChunkedArray)
    assert result.null_count == 0


# ── do_connect OAuth kwargs ──────────────────────────────────────────────────


def test_do_connect_oauth_kwargs(mocker):
    """do_connect passes OAuth kwargs to gizmosql.connect."""
    mock_conn = mocker.MagicMock()
    mock_conn.adbc_get_info.return_value = {"vendor_version": "duckdb v1.1.3"}
    mock_conn.cursor.return_value.__enter__ = mocker.MagicMock()
    mock_conn.cursor.return_value.__exit__ = mocker.MagicMock(return_value=False)

    mock_connect = mocker.patch(
        "xorq.vendor.ibis.backends.gizmosql.gizmosql.connect",
        return_value=mock_conn,
    )

    backend = Backend()
    backend.do_connect(
        host="localhost",
        port=31337,
        auth_type="oauth",
        oauth_port=8080,
        oauth_timeout=120,
        open_browser=False,
        use_encryption=True,
        disable_certificate_verification=True,
    )

    call_kwargs = mock_connect.call_args[1]
    assert call_kwargs["uri"] == "grpc+tls://localhost:31337"
    assert call_kwargs["auth_type"] == "oauth"
    assert call_kwargs["oauth_port"] == 8080
    assert call_kwargs["oauth_timeout"] == 120
    assert call_kwargs["open_browser"] is False
    assert call_kwargs["tls_skip_verify"] is True
    # password auth kwargs should NOT be present for oauth
    assert "username" not in call_kwargs
    assert "password" not in call_kwargs


def test_do_connect_password_kwargs(mocker):
    """do_connect passes username/password for password auth_type."""
    mock_conn = mocker.MagicMock()
    mock_conn.adbc_get_info.return_value = {"vendor_version": "duckdb v1.1.3"}
    mock_conn.cursor.return_value.__enter__ = mocker.MagicMock()
    mock_conn.cursor.return_value.__exit__ = mocker.MagicMock(return_value=False)

    mock_connect = mocker.patch(
        "xorq.vendor.ibis.backends.gizmosql.gizmosql.connect",
        return_value=mock_conn,
    )

    backend = Backend()
    backend.do_connect(
        host="localhost",
        port=31337,
        user="myuser",
        password="mypass",
        auth_type="password",
    )

    call_kwargs = mock_connect.call_args[1]
    assert call_kwargs["uri"] == "grpc://localhost:31337"
    assert call_kwargs["auth_type"] == "password"
    assert call_kwargs["username"] == "myuser"
    assert call_kwargs["password"] == "mypass"
    # OAuth kwargs should NOT be present
    assert "oauth_port" not in call_kwargs


# ── read_xlsx (mocked) ──────────────────────────────────────────────────────


def test_read_xlsx_basic(mocker):
    """read_xlsx calls load_extension, _create_temp_view, and returns a table."""
    backend = Backend()
    mock_table = mocker.MagicMock()
    mocker.patch.object(backend, "load_extension")
    mocker.patch.object(backend, "_create_temp_view")
    mocker.patch.object(backend, "table", return_value=mock_table)

    result = backend.read_xlsx("/tmp/test.xlsx")

    backend.load_extension.assert_called_once_with("excel")
    backend._create_temp_view.assert_called_once()
    backend.table.assert_called_once()
    assert result is mock_table


def test_read_xlsx_with_sheet_and_range(mocker):
    """read_xlsx passes sheet and range kwargs."""
    backend = Backend()
    mock_table = mocker.MagicMock()
    mocker.patch.object(backend, "load_extension")
    mocker.patch.object(backend, "_create_temp_view")
    mocker.patch.object(backend, "table", return_value=mock_table)

    result = backend.read_xlsx("/tmp/test.xlsx", sheet="Sheet2", range="A1:Z10")

    backend.load_extension.assert_called_once_with("excel")
    backend._create_temp_view.assert_called_once()
    assert result is mock_table


# ── to_xlsx (mocked) ────────────────────────────────────────────────────────


def test_to_xlsx(mocker):
    """to_xlsx compiles the expression and executes COPY command."""
    backend = Backend()
    mocker.patch.object(backend, "_run_pre_execute_hooks")
    mocker.patch.object(backend, "compile", return_value="SELECT 1 AS a")
    mocker.patch.object(backend, "load_extension")
    mocker.patch.object(backend, "_execute_ddl")

    mock_expr = mocker.MagicMock()
    backend.to_xlsx(mock_expr, "/tmp/output.xlsx", sheet="MySheet", header=True)

    backend._run_pre_execute_hooks.assert_called_once_with(mock_expr)
    backend.compile.assert_called_once_with(mock_expr, params=None)
    backend.load_extension.assert_called_once_with("excel")
    backend._execute_ddl.assert_called_once()
    # Verify the COPY command format
    copy_cmd = backend._execute_ddl.call_args[0][0]
    assert "COPY" in copy_cmd
    assert "FORMAT 'xlsx'" in copy_cmd
    assert "/tmp/output.xlsx" in copy_cmd


# ── register_filesystem (mocked) ────────────────────────────────────────────


def test_register_filesystem(mocker):
    """register_filesystem delegates to self.con.register_filesystem."""
    backend = Backend()
    mock_con = mocker.MagicMock()
    backend.con = mock_con
    mock_fs = mocker.MagicMock()

    backend.register_filesystem(mock_fs)

    mock_con.register_filesystem.assert_called_once_with(mock_fs)


# ── create_table with temp=True and overwrite=True ──────────────────────────


def test_create_table_temp_overwrite(con, temp_table):
    """create_table with temp=True and overwrite=True uses CREATE + DROP path."""
    t1 = xo.memtable({"x": [1, 2]})
    con.create_table(temp_table, t1, temp=True)
    assert con.table(temp_table).count().execute() == 2

    t2 = xo.memtable({"x": [10, 20, 30]})
    con.create_table(temp_table, t2, overwrite=True, temp=True)
    assert con.table(temp_table).count().execute() == 3


# ── _load_extensions with force_install ─────────────────────────────────────


def test_load_extension_force_install(mocker):
    """load_extension with force_install generates FORCE INSTALL commands."""
    backend = Backend()
    mock_con = mocker.MagicMock()
    backend.con = mock_con

    # Mock the _safe_raw_sql context to return extensions needing install
    mock_cursor = mocker.MagicMock()
    mock_cursor.fetchall.return_value = [
        ("my_ext", False, False),  # not installed, not loaded
    ]
    mock_cursor.__enter__ = mocker.MagicMock(return_value=mock_cursor)
    mock_cursor.__exit__ = mocker.MagicMock(return_value=False)
    mocker.patch.object(backend, "_safe_raw_sql", return_value=mock_cursor)
    mocker.patch.object(backend, "_execute_ddl")

    backend._load_extensions(["my_ext"], force_install=True)

    # Should have called _execute_ddl with FORCE INSTALL and LOAD
    calls = backend._execute_ddl.call_args_list
    assert len(calls) == 2
    assert "FORCE INSTALL 'my_ext'" in calls[0][0][0]
    assert "LOAD 'my_ext'" in calls[1][0][0]


def test_load_extensions_already_loaded(mocker):
    """_load_extensions returns early when all extensions are loaded."""
    backend = Backend()
    mock_con = mocker.MagicMock()
    backend.con = mock_con

    mock_cursor = mocker.MagicMock()
    mock_cursor.fetchall.return_value = []  # nothing to install/load
    mock_cursor.__enter__ = mocker.MagicMock(return_value=mock_cursor)
    mock_cursor.__exit__ = mocker.MagicMock(return_value=False)
    mocker.patch.object(backend, "_safe_raw_sql", return_value=mock_cursor)
    mocker.patch.object(backend, "_execute_ddl")

    backend._load_extensions(["my_ext"])

    # Should NOT have called _execute_ddl
    backend._execute_ddl.assert_not_called()


# ── attach / detach / attach_sqlite (mocked) ────────────────────────────────


def test_attach(mocker):
    """attach delegates to _execute_ddl."""
    backend = Backend()
    mocker.patch.object(backend, "_execute_ddl")

    backend.attach("/tmp/test.duckdb")
    backend._execute_ddl.assert_called_once()
    assert "/tmp/test.duckdb" in backend._execute_ddl.call_args[0][0]


def test_attach_with_name_and_read_only(mocker):
    """attach passes name and read_only options."""
    backend = Backend()
    mocker.patch.object(backend, "_execute_ddl")

    backend.attach("/tmp/test.duckdb", name="mydb", read_only=True)
    cmd = backend._execute_ddl.call_args[0][0]
    assert "ATTACH" in cmd
    assert "mydb" in cmd
    assert "READ_ONLY" in cmd


def test_detach(mocker):
    """detach delegates to _execute_ddl."""
    backend = Backend()
    mocker.patch.object(backend, "_execute_ddl")

    backend.detach("mydb")
    backend._execute_ddl.assert_called_once()
    assert "DETACH" in backend._execute_ddl.call_args[0][0]


def test_attach_sqlite(mocker):
    """attach_sqlite calls load_extension and _execute_ddl."""
    backend = Backend()
    mocker.patch.object(backend, "load_extension")
    mocker.patch.object(backend, "_execute_ddl")

    backend.attach_sqlite("/tmp/test.sqlite")

    backend.load_extension.assert_called_once_with("sqlite")
    assert backend._execute_ddl.call_count == 2
    calls = [c[0][0] for c in backend._execute_ddl.call_args_list]
    assert any("sqlite_all_varchar" in c for c in calls)
    assert any("sqlite_attach" in c for c in calls)


# ── read_mysql (mocked) ─────────────────────────────────────────────────────


def test_read_mysql(mocker):
    """read_mysql delegates to _execute_ddl and returns a table."""
    backend = Backend()
    mock_table = mocker.MagicMock()
    mocker.patch.object(backend, "_load_extensions")
    mocker.patch.object(backend, "_execute_ddl")
    mocker.patch.object(backend, "table", return_value=mock_table)

    result = backend.read_mysql(
        "mysql://root:secret@localhost:3306/shop",
        table_name="orders",
        catalog="my_mysql",
    )

    backend._load_extensions.assert_called_once_with(["mysql"])
    backend._execute_ddl.assert_called_once()
    ddl_cmd = backend._execute_ddl.call_args[0][0]
    assert "ATTACH" in ddl_cmd
    assert "my_mysql" in ddl_cmd
    assert "TYPE mysql" in ddl_cmd
    assert result is mock_table
